from datetime import datetime
from time import mktime

import feedparser
import pytest
import pytz
from django.http import Http404
from django.urls import reverse

from cast.feeds import AtomITunesFeedGenerator, ITunesElements, PodcastFeed
from cast.models import Post


def test_unknown_audio_format():
    pf = PodcastFeed()
    with pytest.raises(Http404):
        pf.set_audio_format("foobar")


class TestFeedCreation:
    pytestmark = pytest.mark.django_db

    def test_add_artwork_true(self, dummy_handler, podcast_with_artwork):
        ie = ITunesElements()
        ie.feed = {"title": "foobar", "link": "bar"}
        ie.add_artwork(podcast_with_artwork, dummy_handler)
        assert "itunes:image" in dummy_handler.aqe
        assert "image" in dummy_handler.se
        assert "image" in dummy_handler.ee

    def test_add_artwork_false(self, dummy_handler, podcast):
        ie = ITunesElements()
        ie.feed = {"title": "foobar", "link": "bar"}
        ie.add_artwork(podcast, dummy_handler)
        assert "itunes:image" not in dummy_handler.aqe
        assert "image" not in dummy_handler.se
        assert "image" not in dummy_handler.ee

    def test_itunes_categories(self, dummy_handler, podcast_with_itunes_categories):
        podcast = podcast_with_itunes_categories
        ie = ITunesElements()
        ie.add_itunes_categories(podcast, dummy_handler)
        assert dummy_handler.se["itunes:category"]["text"] == "foo"
        assert dummy_handler.aqe["itunes:category"][-1]["text"] == "baz"
        assert "itunes:category" in dummy_handler.ee


class TestGeneratedFeeds:
    pytestmark = pytest.mark.django_db

    def test_get_latest_entries_feed(self, client, post, use_dummy_cache_backend):
        feed_url = reverse("cast:latest_entries_feed", kwargs={"slug": post.blog.slug})

        r = client.get(feed_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "xml" in content
        assert post.title in content

    def test_get_podcast_m4a_feed_rss(self, client, episode, use_dummy_cache_backend):
        feed_url = reverse(
            "cast:podcast_feed_rss",
            kwargs={"slug": episode.blog.slug, "audio_format": "m4a"},
        )

        r = client.get(feed_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "rss" in content
        assert episode.title in content

    def test_get_podcast_m4a_feed_atom(self, client, episode):
        feed_url = reverse(
            "cast:podcast_feed_atom",
            kwargs={"slug": episode.blog.slug, "audio_format": "m4a"},
        )

        r = client.get(feed_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "feed" in content
        assert episode.title in content

    def test_podcast_feed_contains_only_podcasts(self, client, post, episode, use_dummy_cache_backend):
        feed_url = reverse(
            "cast:podcast_feed_rss",
            kwargs={"slug": episode.blog.slug, "audio_format": "m4a"},
        )
        r = client.get(feed_url)
        assert r.status_code == 200

        d = feedparser.parse(r.content)
        assert len(d.entries) == 1
        print(episode)
        print(episode.blog)
        print(episode.podcast)
        assert Post.objects.live().descendant_of(episode.blog).count() == 1

    def test_podcast_feed_contains_visible_date_as_pubdate(
        self, client, podcast_episode_with_different_visible_date, use_dummy_cache_backend
    ):
        podcast_episode = podcast_episode_with_different_visible_date
        feed_url = reverse(
            "cast:podcast_feed_rss",
            kwargs={"slug": podcast_episode.podcast.slug, "audio_format": "m4a"},
        )

        r = client.get(feed_url)
        assert r.status_code == 200

        d = feedparser.parse(r.content)
        date_from_feed = datetime.fromtimestamp(mktime(d.entries[0]["published_parsed"]))
        date_from_feed = pytz.utc.localize(date_from_feed)
        assert date_from_feed == podcast_episode.visible_date

    def test_podcast_feed_contains_detail_information(self, client, episode):
        feed_url = reverse(
            "cast:podcast_feed_rss",
            kwargs={"slug": episode.podcast.slug, "audio_format": "m4a"},
        )

        r = client.get(feed_url)
        assert r.status_code == 200

        d = feedparser.parse(r.content)
        content = d.entries[0]["content"][0]["value"]
        assert "in_all" in content
        assert "only_in_detail" in content


def test_itunes_elements_add_root_elements_index_error(mocker):
    class MockedHandler:
        def addQuickElement(self, name, content=None, attrs=None):
            if name == "lastBuildDate":
                raise IndexError

        def startElement(self, name, attrs):
            pass

        def endElement(self, name):
            pass

    mocker.patch("cast.feeds.ITunesElements.add_artwork")
    mocker.patch("cast.feeds.rfc2822_date")
    atom_itunes_feed_generator = AtomITunesFeedGenerator("title", "link", "description")
    atom_itunes_feed_generator.feed = mocker.MagicMock()
    handler = MockedHandler()
    add_returned = atom_itunes_feed_generator.add_root_elements(handler)
    assert add_returned is None


def test_itunes_elements_add_item_elements_post_block(mocker):
    mocker.patch("cast.feeds.Atom1Feed.add_item_elements")
    post = mocker.MagicMock()
    post.block = True
    handler = mocker.MagicMock()
    atom_itunes_feed_generator = AtomITunesFeedGenerator("title", "link", "description")
    atom_itunes_feed_generator.add_item_elements(handler, {"post": post})
    handler.addQuickElement.assert_called_with("itunes:block", "yes")


def test_podcast_feed_categories_and_keywords():
    class MockedBlog:
        categories = True
        keywords = "foo, bar, baz"
        itunes_categories = "one,two,three"

    podcast_feed = PodcastFeed()

    blog = MockedBlog()
    # test categories -> first category
    categories = podcast_feed.categories(blog)
    assert categories == ("foo",)

    # itunes_categories -> split itunes_categories
    assert podcast_feed.itunes_categories(blog) == blog.itunes_categories.split(",")

    # item_keywords -> item.keywords
    assert podcast_feed.item_keywords(blog) == blog.keywords



import json
import random
import re
from collections import defaultdict
from collections.abc import Callable, Iterator
from typing import Any, Optional

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from model_utils.models import TimeStampedModel

token_pattern = re.compile(r"(?u)\b\w\w+\b")
standard_tokenizer = token_pattern.findall


def regex_tokenize(message: str) -> list[str]:
    return standard_tokenizer(message.lower())


Probabilities = dict[str, float]
Counts = dict[str, int]


def normalize(probabilities: Probabilities) -> Probabilities:
    try:
        factor = 1.0 / float(sum(probabilities.values()))
    except ZeroDivisionError:
        # not possible to scale -> skip
        return probabilities
    for name, value in probabilities.items():
        probabilities[name] *= factor
    return probabilities


Messages = list[tuple[str, str]]


class NaiveBayes:
    def __init__(
        self,
        tokenize: Callable = regex_tokenize,
        prior_probabilities: Optional[Probabilities] = None,
        word_label_counts: Optional[dict[str, Counts]] = None,
    ):
        self.tokenize = tokenize
        if prior_probabilities is None:
            prior_probabilities = {}
        self.prior_probabilities = prior_probabilities
        if word_label_counts is None:
            self.word_label_counts: dict[str, Counts] = defaultdict(lambda: defaultdict(int))
        else:
            self.word_label_counts = word_label_counts
        self.number_of_words = self.get_number_of_words(self.word_label_counts)
        self.number_of_all_words = sum(self.number_of_words.values())

    @staticmethod
    def get_label_counts(messages: Messages) -> Counts:
        label_counts: Counts = defaultdict(int)
        for label, text in messages:
            label_counts[label] += 1
        return label_counts

    def set_prior_probabilities(self, label_counts: Counts) -> None:
        number_of_messages = sum(label_counts.values())
        self.prior_probabilities = {label: count / number_of_messages for label, count in label_counts.items()}

    def set_word_label_counts(self, messages: Messages) -> None:
        counts = self.word_label_counts
        for label, text in messages:
            for word in self.tokenize(text):
                counts[word][label] += 1

    @staticmethod
    def get_number_of_words(word_label_counts: dict[str, Counts]) -> Counts:
        number_of_words: Counts = defaultdict(int)
        for word, counts in word_label_counts.items():
            for label, count in counts.items():
                number_of_words[label] += 1
        return number_of_words

    def fit(self, messages: Messages) -> "NaiveBayes":
        self.set_prior_probabilities(self.get_label_counts(messages))
        self.set_word_label_counts(messages)
        self.number_of_words = self.get_number_of_words(self.word_label_counts)
        self.number_of_all_words = sum(self.number_of_words.values())
        return self

    @staticmethod
    def update_probabilities(probabilities: Probabilities, counts_per_label: Counts, number_of_all_words: int):
        updated_probabilities = {}
        for label, prior_probability in probabilities.items():
            word_count = counts_per_label.get(label, 0.5)
            word_probability = word_count / number_of_all_words
            updated_probabilities[label] = prior_probability * word_probability
        return updated_probabilities

    def predict(self, message: str) -> Probabilities:
        probabilities = dict(self.prior_probabilities)
        for word in self.tokenize(message):
            counts_per_label = self.word_label_counts.get(word, {})
            probabilities = normalize(
                self.update_probabilities(probabilities, counts_per_label, self.number_of_all_words)
            )
        return probabilities

    def predict_label(self, message: str) -> Optional[str]:
        probabilities = self.predict(message)
        if len(probabilities) == 0:
            return None
        return sorted(((prob, label) for label, prob in probabilities.items()), reverse=True)[0][1]

    def dict(self):
        return {
            "class": "NaiveBayes",
            "prior_probabilities": self.prior_probabilities,
            "word_label_counts": self.word_label_counts,
        }

    def __eq__(self, other: Any) -> bool:
        return (
            self.prior_probabilities == other.prior_probabilities and self.word_label_counts == other.word_label_counts
        )


class ModelEncoder(DjangoJSONEncoder):
    def default(self, obj: Any) -> dict:
        if isinstance(obj, NaiveBayes):
            return obj.dict()
        return super().default(obj)


class ModelDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        kwargs["object_hook"] = self.model_decode
        super().__init__(*args, **kwargs)

    @staticmethod
    def model_decode(obj):
        if obj.get("class") == "NaiveBayes":
            del obj["class"]
            return NaiveBayes(**obj)
        return obj


Performance = dict[str, float]


class Evaluation:
    """Simple cross validation evaluation."""

    def __init__(self, model_class: type = NaiveBayes, num_folds: int = 3):
        self.model_class = model_class
        self.num_folds = num_folds

    @staticmethod
    def split_into_labels(messages: Messages) -> dict[str, Messages]:
        """Split messages into a dict of labels and labeled messages."""
        data_per_label = defaultdict(list)
        for label, text in messages:
            data_per_label[label].append((label, text))
        return data_per_label

    @staticmethod
    def split_into_folds(messages: Messages, num_folds: int) -> list[Messages]:
        """
        Split the messages into num_folds folds.
        """
        random.shuffle(messages)
        fold_size = len(messages) // num_folds
        folds = []
        for i in range(num_folds):
            folds.append(messages[i * fold_size : (i + 1) * fold_size])  # noqa: E203
        return folds

    def stratified_split_into_folds(self, messages: Messages) -> list[Messages]:
        """Split the messages in a stratified way into num_folds folds."""
        data_per_label = self.split_into_labels(messages)
        labeled_folds = []
        for label, labeled_messages in data_per_label.items():
            labeled_folds.append(self.split_into_folds(labeled_messages, self.num_folds))
        folds = []
        for nested_fold in zip(*labeled_folds):
            fold = []
            for labeled_fold in nested_fold:
                fold.extend(labeled_fold)
            folds.append(fold)
        return folds

    @staticmethod
    def generate_train_test(folds: list[Messages]) -> Iterator[tuple[Messages, Messages]]:
        """From a list of n cross-validation folds, generate n train and test sets."""
        for i, n in enumerate(folds):
            all_but_n = []
            for j, m in enumerate(folds):
                if i != j:
                    all_but_n.extend(m)
            yield all_but_n, n

    @staticmethod
    def evaluate_model(model: Any, test_messages: Messages) -> dict[str, Counts]:
        """Build a confusion matrix for the model on the test messages."""
        outcomes = (("true", "false"), ("positive", "negative"))
        possible_results = [f"{a}_{b}" for b in outcomes[1] for a in outcomes[0]]
        result_template = dict.fromkeys(possible_results, 0)

        labels = set(model.prior_probabilities)
        label_results = {label: dict(result_template) for label in labels}

        for label, message in test_messages:
            predicted = model.predict_label(message)
            if label == predicted:
                label_results[label]["true_positive"] += 1
            else:
                label_results[label]["false_negative"] += 1
                label_results[predicted]["false_positive"] += 1
        return label_results

    @staticmethod
    def get_precision_recall_f1(result: Counts) -> tuple[float, float, float]:
        """Actual implementation of precision, recall and f1 from tp, fp, fn."""
        tp = result["true_positive"]
        fp = result["false_positive"]
        fn = result["false_negative"]
        precision = tp / (tp + fp) if tp + fp > 0 else 0
        recall = tp / (tp + fn) if tp + fn > 0 else 0
        try:
            f1 = 2 * precision * recall / (precision + recall)
        except ZeroDivisionError:
            f1 = 0
        return precision, recall, f1

    def calc_performance(self, results: dict[str, Counts]) -> dict[str, Performance]:
        """Calc precision, recall and f1 for each label."""
        performance = {}
        print("results: ", results)
        for label, result in results.items():
            precision, recall, f1 = self.get_precision_recall_f1(result)
            performance[label] = {
                "precision": precision,
                "recall": recall,
                "f1": f1,
            }
        return performance

    def evaluate(self, messages: Messages) -> dict[str, Performance]:
        """
        Evaluate the model on the given messages. Use stratified cross validation
        to determine precision, recall and f1 score.
        """
        folds = self.stratified_split_into_folds(messages)
        results = None
        for train, test in self.generate_train_test(folds):
            model = self.model_class().fit(train)
            results = self.evaluate_model(model, test)
        if results is None:
            raise ValueError("No results")
        return self.calc_performance(results)


class SpamFilter(TimeStampedModel):
    """
    A Django model that stores a trained spam filter.

    The model itself is stored in the model JSONField. There's a second JSONField
    where some performance indicators like precision, recall and f1 are stored.

    There are some helper methods to generate training data from comments and
    retrain the model from scratch for example when a significant amount of
    new training data was added.
    """

    name = models.CharField(unique=True, max_length=128)
    model = models.JSONField(verbose_name="Spamfilter Model", default=dict, encoder=ModelEncoder, decoder=ModelDecoder)
    performance = models.JSONField(verbose_name="Spamfilter Performance Indicators", default=dict)

    @classmethod
    def comment_to_message(cls, comment: Any) -> str:
        return f"{comment.name} {comment.email} {comment.title} {comment.comment}"

    @classmethod
    def get_training_data_comments(cls) -> Messages:
        """
        Keep this as a classmethod in SpamFilter to make it available for all code importing SpamFilter.
        """
        from django_comments import get_model as get_comments_model

        comment_class = get_comments_model()
        train = []
        for comment in comment_class.objects.all():
            label = "ham" if (comment.is_public and not comment.is_removed) else "spam"
            message = cls.comment_to_message(comment)
            train.append((label, message))
        return train

    def retrain_from_scratch(self, train: Messages) -> None:
        """
        Retrain on all comments for now. Later on there might be
        different spam filters for different blogs/sites..
        """
        model = NaiveBayes().fit(train)
        self.model = model
        self.performance = Evaluation().evaluate(train)
        self.save()

    @classmethod
    def get_default(cls) -> "SpamFilter":
        return cls.objects.first()



import environ

ROOT_DIR = environ.Path(__file__) - 1
APPS_DIR = ROOT_DIR.path("cast")

DEBUG = False
USE_TZ = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "mbmcf(_0(y@^nlf6w#1nq%s7&nzcfvx#ok$iwu8)i^d+^96h*="

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "tests/test_database.sqlite3",
        # if this is not set, an in memory database is used
        # for tests by default _get_test_db_name
        "TEST": {"NAME": "tests/test_database.sqlite3"},
    },
}

ROOT_URLCONF = "tests.urls"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.admin",
    "django.contrib.messages",
    "django_filters",
    "django_vite",
    "crispy_forms",  # Form layouts
    "crispy_bootstrap4",
    "fluent_comments",  # must be before django_comments
    "threadedcomments",
    "django_comments",
    "modelcluster",
    "taggit",
    "wagtail.contrib.forms",
    "wagtail.contrib.settings",
    "wagtail.contrib.redirects",
    "wagtail.embeds",
    "wagtail.sites",
    "wagtail.users",
    "wagtail.snippets",
    "wagtail.documents",
    "wagtail.images",
    "wagtail.search",
    "wagtail.admin",
    "wagtail",
    "cast.apps.CastConfig",
]

SITE_ID = 1

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "wagtail.contrib.redirects.middleware.RedirectMiddleware",
    "django_htmx.middleware.HtmxMiddleware",
]

# only gets invoked for python manage.py test
TEST_RUNNER = "cast.runner.PytestTestRunner"

# STATIC FILE CONFIGURATION
# ------------------------------------------------------------------------------
# See: https://docs.djangoproject.com/en/dev/ref/settings/#static-root
STATIC_ROOT = str(ROOT_DIR("staticfiles"))

# See: https://docs.djangoproject.com/en/dev/ref/settings/#static-url
STATIC_URL = "/static/"

# See: https://docs.djangoproject.com/en/dev/ref/contrib/staticfiles/#std:setting-STATICFILES_DIRS
STATICFILES_DIRS = [str(APPS_DIR.path("static"))]

# See: https://docs.djangoproject.com/en/dev/ref/contrib/staticfiles/#staticfiles-finders
STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
]

# MEDIA CONFIGURATION
# ------------------------------------------------------------------------------
# See: https://docs.djangoproject.com/en/dev/ref/settings/#media-root
MEDIA_ROOT = str(ROOT_DIR("media"))

# See: https://docs.djangoproject.com/en/dev/ref/settings/#media-url
MEDIA_URL = "/media/"


# TEMPLATE CONFIGURATION
# ------------------------------------------------------------------------------
# See: https://docs.djangoproject.com/en/dev/ref/settings/#templates
TEMPLATES = [
    {
        # See: https://docs.djangoproject.com/en/dev/ref/settings/#std:setting-TEMPLATES-BACKEND
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        # See: https://docs.djangoproject.com/en/dev/ref/settings/#template-dirs
        "DIRS": [str(APPS_DIR.path("templates"))],
        "OPTIONS": {
            # See: https://docs.djangoproject.com/en/dev/ref/settings/#template-debug
            "debug": DEBUG,
            # See: https://docs.djangoproject.com/en/dev/ref/settings/#template-loaders
            # https://docs.djangoproject.com/en/dev/ref/templates/api/#loader-types
            "loaders": [
                "django.template.loaders.filesystem.Loader",
                "django.template.loaders.app_directories.Loader",
            ],
            # See: https://docs.djangoproject.com/en/dev/ref/settings/#template-context-processors
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                # Your stuff: custom template context processors go here
            ],
        },
    }
]

# TEMPLATE LOADERS
# ------------------------------------------------------------------------------
# Keep templates in memory so tests run faster
TEMPLATES[0]["OPTIONS"]["loaders"] = [  # type: ignore
    [
        "django.template.loaders.cached.Loader",
        [
            "django.template.loaders.filesystem.Loader",
            "django.template.loaders.app_directories.Loader",
        ],
    ]
]

# Comments
COMMENTS_APP = "fluent_comments"

# For Django >= 3.2 - auto field
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# This provided a 70% speedup
# use a faster password hasher - don't do this in production, it's only for testing
PASSWORD_HASHERS = [
    "django.contrib.auth.hashers.MD5PasswordHasher",
]

# Use in memory cache for tests
CACHES = {"default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"}}

# Wagtail
WAGTAILADMIN_BASE_URL = "/cms"
WAGTAILIMAGES_MAX_UPLOAD_SIZE = 30 * 1024 * 1024

# crispy forms
CRISPY_TEMPLATE_PACK = "bootstrap4"
CRISPY_ALLOWED_TEMPLATE_PACKS = "bootstrap4"

# view handling csrf failures
CSRF_FAILURE_VIEW = "cast.views.defaults.csrf_failure"

# django_vite
DJANGO_VITE = {
    "cast": {
        "dev_mode": True,
    }
}



# Generated by Django 4.1.4 on 2023-02-04 16:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0035_remove_new_prefix_podcast_fields"),
    ]

    operations = [
        migrations.AlterField(
            model_name="blog",
            name="author",
            field=models.CharField(
                blank=True,
                default=None,
                help_text="Freeform text that will be used in the feed.",
                max_length=255,
                null=True,
            ),
        ),
    ]



from django.http import HttpRequest
from django.urls import include, path, reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from taggit.models import Tag
from wagtail import hooks
from wagtail.admin.menu import MenuItem
from wagtail.admin.panels import FieldPanel
from wagtail.permission_policies.collections import CollectionOwnershipPermissionPolicy
from wagtail.snippets.models import register_snippet
from wagtail.snippets.views.snippets import SnippetViewSet

from .admin_urls import audio, video
from .models import Audio, Video


@hooks.register("register_admin_urls")
def register_admin_urls() -> list:
    return [
        path("audio/", include((audio, "castaudio"), namespace="castaudio")),
        path("media/", include((video, "castvideo"), namespace="castvideo")),
    ]


class VideoMenuItem(MenuItem):
    def is_shown(self, request: HttpRequest) -> bool:
        permission_policy = CollectionOwnershipPermissionPolicy(Video, auth_model=Video, owner_field_name="user")
        return permission_policy.user_has_any_permission(request.user, ["add", "change", "delete"])


@hooks.register("register_admin_menu_item")
def register_video_menu_item() -> VideoMenuItem:
    return VideoMenuItem(
        _("Video"),
        reverse("castvideo:index"),
        name="video",
        classnames="icon icon-media",
        order=300,
    )


@hooks.register("insert_editor_js")
def editor_js() -> str:
    return format_html(
        """
        <script>
            window.chooserUrls.videoChooser = '{0}';
        </script>
        """,
        reverse("castvideo:chooser"),
    )


class AudioMenuItem(MenuItem):
    def is_shown(self, request: HttpRequest) -> bool:
        permission_policy = CollectionOwnershipPermissionPolicy(Audio, auth_model=Audio, owner_field_name="user")
        return permission_policy.user_has_any_permission(request.user, ["add", "change", "delete"])


@hooks.register("register_admin_menu_item")
def register_audio_menu_item() -> AudioMenuItem:
    return AudioMenuItem(
        _("Audio"),
        reverse("castaudio:index"),
        name="audio",
        classnames="icon icon-media",
        order=300,
    )


@hooks.register("insert_editor_js")
def editor_js_audio() -> str:
    return format_html(
        """
        <script>
            window.chooserUrls.audioChooser = '{0}';
        </script>
        """,
        reverse("castaudio:chooser"),
    )


class TagsSnippetViewSet(SnippetViewSet):
    """
    This is a snippet viewset for the taggit Tag model. It is used to
    be able to do crud operations on tags from the admin menu.
    """

    panels = [FieldPanel("name")]  # only show the name field
    model = Tag
    icon = "tag"  # change as required
    add_to_admin_menu = True
    menu_label = "Tags"
    menu_order = 200  # will put in 3rd place (000 being 1st, 100 2nd)
    list_display = ["name", "slug"]
    search_fields = ("name",)


register_snippet(TagsSnippetViewSet)



from unittest.mock import patch

import pytest
from django.urls import reverse
from django_comments import get_model as get_comments_model
from django_comments import signals
from django_comments.forms import CommentForm

from cast.moderation import Moderator


class TestPostComments:
    pytestmark = pytest.mark.django_db

    def test_comment_form_not_included(self, client, post, comments_not_enabled):
        detail_url = post.get_url()
        r = client.get(detail_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert post.title in content
        assert "form" not in r.context

    def test_comment_form_included(self, client, post, comments_enabled):
        detail_url = post.get_url()
        r = client.get(detail_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert post.title in content
        assert isinstance(r.context["form"], CommentForm)

    def test_comment_in_comment_list(self, client, post, comment, comments_enabled):
        detail_url = post.get_url()
        r = client.get(detail_url)
        assert r.status_code == 200

        assert comment in r.context["comment_list"]

    def test_add_new_comment(self, client, post, comments_enabled):
        ajax_url = reverse("comments-post-comment-ajax")
        detail_url = post.get_url()

        r = client.get(detail_url)
        content = r.content.decode("utf-8")
        security_hash, timestamp = None, None
        for line in content.split("\n"):
            if "security_hash" in line:
                for part in line.split("input"):
                    if "security_hash" in part:
                        for attr in part.split(" "):
                            if "value" in attr:
                                security_hash = attr.split('"')[1]
                    if "timestamp" in part:
                        for attr in part.split():
                            if "value" in attr:
                                timestamp = attr.split('"')[1]

        data = {
            "content_type": "cast.post",
            "object_pk": str(post.pk),
            "comment": "new content",
            "name": "Name",
            "email": "fuz@baz.com",
            "title": "buzz",
            "security_hash": security_hash,
            "timestamp": timestamp,
        }

        r = client.post(ajax_url, data, HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        assert r.status_code == 200

        rdata = r.json()
        assert rdata["success"]

        comment = get_comments_model().objects.get(pk=rdata["comment_id"])
        assert comment.comment == data["comment"]


class TestCommentModeration:
    @classmethod
    def setup_class(cls):
        class Stub:
            pass

        class StubComment:
            name = "name"
            email = "foo@example.com"
            title = "foobar title"
            comment = "some comment"
            is_removed = False
            is_public = True

        class PredictSpam:
            @staticmethod
            def predict_label(_):
                return "spam"

        class PredictHam:
            @staticmethod
            def predict_label(_):
                return "ham"

        class SpamFilter:
            def __init__(self, model):
                self.model = model

        cls.post = post = Stub()
        cls.stub_class = Stub
        cls.comment_class = StubComment
        cls.comment = comment = StubComment()
        comment.content_object = post
        cls.request = Stub()
        cls.predict_spam = SpamFilter(PredictSpam())
        cls.predict_ham = SpamFilter(PredictHam())

    @pytest.mark.django_db
    def test_spamfilter_is_none(self):
        with patch("fluent_comments.receivers.default_moderator", new=Moderator(self.stub_class, spamfilter=None)):
            signals.comment_will_be_posted.send(sender=self.comment_class, comment=self.comment, request=self.request)
            assert self.comment.is_public
            assert not self.comment.is_removed

    def test_moderated_comment_marked_is_removed(self):
        with patch(
            "fluent_comments.receivers.default_moderator", new=Moderator(self.stub_class, spamfilter=self.predict_spam)
        ):
            signals.comment_will_be_posted.send(sender=self.comment_class, comment=self.comment, request=self.request)
            assert self.comment.is_removed
            assert not self.comment.is_public

    def test_moderated_comment_is_not_marked_is_removed(self):
        with patch(
            "fluent_comments.receivers.default_moderator", new=Moderator(self.stub_class, spamfilter=self.predict_ham)
        ):
            signals.comment_will_be_posted.send(sender=self.comment_class, comment=self.comment, request=self.request)
            assert self.comment.is_public
            assert not self.comment.is_removed



import logging

from django.forms import ModelForm
from django.http import HttpRequest, HttpResponse

logger = logging.getLogger(__name__)


class FileUploadResponseMixin:
    @staticmethod
    def get_success_url() -> None:
        return None

    def form_valid(self, form: ModelForm) -> HttpResponse:
        model = form.save(commit=False)
        super().form_valid(form)  # type: ignore
        return HttpResponse(f"{model.pk}", status=201)


class AddRequestUserMixin:
    request: HttpRequest
    user_field_name = "user"

    def form_valid(self, form: ModelForm) -> bool:
        model = form.save(commit=False)
        setattr(model, self.user_field_name, self.request.user)
        return super().form_valid(form)  # type: ignore



# Generated by Django 4.1.6 on 2023-02-07 15:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("cast", "0037_alter_episode_block_alter_episode_explicit_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="episode",
            name="keywords",
            field=models.CharField(
                blank=True,
                default="",
                help_text="A comma-delimited-list of up to 12 words for iTunes\n            searches. Perhaps include misspellings of the title.",
                max_length=255,
            ),
        ),
    ]



import pytest

from cast import appsettings
from cast.models.index_pages import Blog


class TestBlogIndex:
    pytestmark = pytest.mark.django_db

    def test_post_in_blog_index(self, client, post):
        blog_url = post.blog.get_url()

        r = client.get(blog_url)
        assert r.status_code == 200

        assert post in r.context["posts"]

    def test_unpublished_post_not_in_blog_index(self, client, unpublished_post):
        blog_url = unpublished_post.blog.get_url()

        r = client.get(blog_url)
        assert r.status_code == 200

        assert unpublished_post not in r.context["posts"]

    def test_post_overview_content_in_blog_index_but_not_detail(self, client, post):
        blog_url = post.blog.get_url()

        r = client.get(blog_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "in_all" in content
        assert "only_in_detail" not in content

    def test_post_in_blog_noindex_flag(self, client, post):
        # Set the noindex flag for the blog
        blog = post.blog
        blog.noindex = True
        blog.save()

        # Make sure the blog index page contains the noindex meta tag
        blog_url = post.blog.get_url()
        r = client.get(blog_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert '<meta name="robots" content="noindex">' in content

        # Make sure the post detail page contains the noindex meta tag
        post_url = post.get_url()
        r = client.get(post_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert '<meta name="robots" content="noindex">' in content

    def test_blog_template_base_dir_overwrites_site_setting(self, simple_request):
        blog = Blog(template_base_dir="plain")
        chosen_base_dir = "foobar"
        blog.template_base_dir = chosen_base_dir
        template = blog.get_template(simple_request)
        assert chosen_base_dir in template

    def test_session_template_base_dir_overwrites_blog_setting(self, simple_request):
        blog = Blog(template_base_dir="plain")
        chosen_base_dir = "from_session"
        simple_request.session["template_base_dir"] = "from_session"
        template = blog.get_template(simple_request)
        assert chosen_base_dir in template

    def test_blog_use_partial_template_for_htmx_request_without_target(self, caplog, htmx_request_without_target):
        Blog().get_template(htmx_request_without_target)
        assert "HTMX target is None" in caplog.text

    def test_blog_use_partial_template_for_htmx_request(self, htmx_request):
        blog = Blog()
        template = blog.get_template(htmx_request)
        assert "_list_of_posts_and_paging_controls.html" in template

    def test_post_in_blog_inherits_template_base_dir(self, post):
        blog = post.blog
        chosen_base_dir = "foobar"
        blog.template_base_dir = chosen_base_dir
        template = post.get_template(None)
        assert chosen_base_dir in template

    def test_template_base_dir_in_context(self, simple_request):
        # from blog model
        blog = Blog(template_base_dir="plain")
        context = blog.get_context(simple_request)
        assert context["template_base_dir"] == "plain"

        # from session
        chosen_base_dir = "from_session"
        simple_request.session["template_base_dir"] = "from_session"
        context = blog.get_context(simple_request)
        assert context["template_base_dir"] == chosen_base_dir


class TestBlogIndexFilter:
    pytestmark = pytest.mark.django_db

    def test_date_facet_filter_shown(self, client, post_with_date):
        blog_url = post_with_date.blog.get_url()
        r = client.get(blog_url)
        assert r.status_code == 200

        date_to_find = post_with_date.visible_date.strftime("%Y-%m")
        content = r.content.decode("utf-8")
        assert "html" in content
        assert date_to_find in content

    def test_date_facet_filter_shown_exclusively(self, client, post_with_date, post_with_different_date):
        blog_url = post_with_date.blog.get_url()
        r = client.get(blog_url)
        assert r.status_code == 200

        # assert both date facets are shown
        date_to_find = post_with_date.visible_date.strftime("%Y-%m")
        different_date_to_find = post_with_different_date.visible_date.strftime("%Y-%m")
        content = r.content.decode("utf-8")
        assert "html" in content
        assert date_to_find in content
        assert different_date_to_find in content

        # assert only one date facet is shown if one is selected
        blog_url = f"{blog_url}?date_facets={date_to_find}"

        r = client.get(blog_url)
        assert r.status_code == 200
        content = r.content.decode("utf-8")
        assert "html" in content
        assert date_to_find in content
        assert different_date_to_find not in content  # attention

    def test_selecting_non_existing_date_facet_filters_all_posts(self, client, post_with_date):
        blog_url = post_with_date.blog.get_url()
        blog_url = f"{blog_url}?date_facets=1999-01"
        r = client.get(blog_url)
        assert r.status_code == 200
        assert len(r.context["posts"]) == 0

    def test_date_facet_garbage(self, client, post_with_date):
        blog_url = post_with_date.blog.get_url()
        blog_url = f"{blog_url}?date_facets=garbage"
        r = client.get(blog_url)
        assert r.status_code == 200

        date_to_find = post_with_date.visible_date.strftime("%Y-%m")
        content = r.content.decode("utf-8")
        assert "html" in content
        assert date_to_find in content


class TestBlogIndexSearch:
    pytestmark = pytest.mark.django_db

    def test_fulltext_search_all(self, client, post, post_with_search):
        blog_url = post.blog.get_url()
        r = client.get(blog_url)
        assert r.status_code == 200

        # assert initially both posts are in the post list
        assert len(r.context["posts"]) == 2

    def test_fulltext_search_title(self, client, post, post_with_search):
        blog_url = post.blog.get_url()
        blog_url_title = f"{blog_url}?search={post_with_search.title}"
        r = client.get(blog_url_title)
        assert r.status_code == 200

        # assert search by title only yields post_with search
        posts = r.context["posts"]
        assert len(posts) == 1
        assert posts[0].pk == post_with_search.pk

    def test_fulltext_search_body(self, client, post, post_with_search):
        blog_url = post.blog.get_url()
        blog_url_content = f"{blog_url}?search={post_with_search.query}"
        r = client.get(blog_url_content)
        assert r.status_code == 200

        # assert search by title only yields post_with search
        posts = r.context["posts"]
        assert len(posts) == 1
        assert posts[0].pk == post_with_search.pk


@pytest.fixture()
def post_list_paginate_by_1():
    previous = appsettings.POST_LIST_PAGINATION
    appsettings.POST_LIST_PAGINATION = 1
    yield appsettings.POST_LIST_PAGINATION
    appsettings.POST_LIST_PAGINATION = previous


class TestBlogIndexPagination:
    pytestmark = pytest.mark.django_db

    def test_one_post_is_not_paginated(self, client, post):
        blog_url = post.blog.get_url()
        r = client.get(blog_url)
        assert r.status_code == 200

        # make sure list of posts is not paginated
        assert r.context["is_paginated"] is False

    def test_two_posts_are_paginated_with_pagination_1(self, client, post, post_with_search, post_list_paginate_by_1):
        blog_url = post.blog.get_url()
        r = client.get(blog_url)
        assert r.status_code == 200

        # make sure posts queryset is paginated and len(posts) == 1
        assert r.context["is_paginated"] is True
        assert len(r.context["posts"]) == 1

    def test_invalid_page_param(self, client, post):
        blog_url = post.blog.get_url()
        blog_url_with_invalid_page = f"{blog_url}?page=foo"
        r = client.get(blog_url_with_invalid_page)
        assert r.status_code == 404

    def test_page_param_last(self, client, post):
        blog_url = post.blog.get_url()
        blog_url_with_last_page = f"{blog_url}?page=last"
        r = client.get(blog_url_with_last_page)
        assert r.status_code == 200

    def test_invalid_page_number(self, client, post):
        blog_url = post.blog.get_url()
        blog_url_with_invalid_pagenum = f"{blog_url}?page=666"
        r = client.get(blog_url_with_invalid_pagenum)
        assert r.status_code == 404



# Generated by Django 4.2.4 on 2023-08-20 12:57

from django.db import migrations, models
import django.db.models.deletion
import modelcluster.contrib.taggit
import modelcluster.fields


class Migration(migrations.Migration):
    dependencies = [
        ("taggit", "0005_auto_20220424_2025"),
        ("cast", "0049_added_category_snippets"),
    ]

    operations = [
        migrations.CreateModel(
            name="PostTag",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "content_object",
                    modelcluster.fields.ParentalKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="tagged_items", to="cast.post"
                    ),
                ),
                (
                    "tag",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(app_label)s_%(class)s_items",
                        to="taggit.tag",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="post",
            name="tags",
            field=modelcluster.contrib.taggit.ClusterTaggableManager(
                blank=True,
                help_text="A comma-separated list of tags.",
                through="cast.PostTag",
                to="taggit.Tag",
                verbose_name="tags",
            ),
        ),
    ]



# Generated by Django 4.1.4 on 2023-02-04 07:28

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0032_remove_timestampedmodel_because_wagtail"),
    ]

    operations = [
        migrations.CreateModel(
            name="Podcast",
            fields=[
                (
                    "blog_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="cast.blog",
                    ),
                ),
                (
                    "new_itunes_categories",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="A json dict of itunes categories pointing to lists of subcategories. Taken from this list https://validator.w3.org/feed/docs/error/InvalidItunesCategory.html",
                        max_length=512,
                        verbose_name="itunes_categories",
                    ),
                ),
                (
                    "new_keywords",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="A comma-delimitedlist of up to 12 words for iTunes\n            searches. Perhaps include misspellings of the title.",
                        max_length=255,
                        verbose_name="keywords",
                    ),
                ),
                (
                    "new_explicit",
                    models.PositiveSmallIntegerField(
                        choices=[(1, "yes"), (2, "no"), (3, "clean")],
                        default=1,
                        help_text="``Clean`` will put the clean iTunes graphic by it.",
                        verbose_name="explicit",
                    ),
                ),
                (
                    "new_itunes_artwork",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="cast.itunesartwork"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("cast.blog",),
        ),
    ]



# Generated by Django 4.2.4 on 2023-08-15 11:58

from django.db import migrations, models
import modelcluster.fields


class Migration(migrations.Migration):
    dependencies = [
        ("cast", "0048_added_visible_date_index_for_wagtail_api"),
    ]

    operations = [
        migrations.CreateModel(
            name="PostCategory",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="The name for this category", max_length=255, unique=True)),
                (
                    "slug",
                    models.SlugField(
                        help_text="A slug to identify posts by this category", unique=True, verbose_name="slug"
                    ),
                ),
            ],
            options={
                "verbose_name": "Post Category",
                "verbose_name_plural": "Post Categories",
                "ordering": ["name"],
            },
        ),
        migrations.AddField(
            model_name="post",
            name="categories",
            field=modelcluster.fields.ParentalManyToManyField(blank=True, to="cast.postcategory"),
        ),
    ]



# Generated by Django 4.1.4 on 2023-02-04 13:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0033_add_new_podcast_model"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="blog",
            name="explicit",
        ),
        migrations.RemoveField(
            model_name="blog",
            name="itunes_artwork",
        ),
        migrations.RemoveField(
            model_name="blog",
            name="itunes_categories",
        ),
        migrations.RemoveField(
            model_name="blog",
            name="keywords",
        ),
    ]



from django.core.management.base import BaseCommand
from wagtail.images.models import Image

from ...models import File, Video
from ...utils import storage_walk_paths
from .storage_backend import get_production_and_backup_storage


class Command(BaseCommand):
    help = (
        "show media files which are in the production storage backend, but not in database and optionally delete them "
        "(requires Django >= 4.2 and production and backup storage configured)"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--delete",
            action="store_true",
            dest="delete",
            default=False,
            help="Delete stale files instead of only showing them",
        )

    @staticmethod
    def get_paths(storage):
        paths = {}
        for num, path in enumerate(storage_walk_paths(storage)):
            size = storage.size(path)
            paths[path] = size
            if num % 100 == 0:
                print(".", end="", flush=True)
            # print(path, size / 2 ** 20)
        return paths

    @staticmethod
    def get_image_paths() -> set[str]:
        paths = set()
        for image in Image.objects.all():
            paths.add(image.file.name)
            for rendition in image.renditions.all():
                paths.add(rendition.file.name)
        return paths

    def get_models_paths(self):
        paths = self.get_image_paths()
        for video in Video.objects.all():
            for path in video.get_all_paths():
                paths.add(path)
        for misc_file in File.objects.all():
            for path in misc_file.get_all_paths():
                paths.add(path)
        return paths

    def handle(self, *args, **options):
        production_storage, _ = get_production_and_backup_storage()
        paths_from_models = self.get_models_paths()

        print("stale production")
        production_paths = self.get_paths(production_storage)
        stale_production = {}
        for path, size in production_paths.items():
            if path not in paths_from_models:
                print(path)
                stale_production[path] = size
        print(f"stale production size: {sum(stale_production.values()) / 2 ** 20} Mb")

        if options["delete"]:
            for path in stale_production.keys():
                production_storage.delete(path)



import pytest

from cast.renditions import (
    Height,
    ImageFormat,
    Rectangle,
    RenditionFilter,
    RenditionFilters,
    Width,
    calculate_fitting_width,
    get_image_format_by_name,
    get_rendition_filters_for_image_and_slot,
)

rect = Rectangle
w1, w53, w106, w120, w159, w1110, w4000, w6000, w8000 = (
    Width(x) for x in [1, 53, 106, 120, 159, 1110, 4000, 6000, 8000]
)
h1, h80, h740, h4000, h6000 = Height(1), Height(80), Height(740), Height(4000), Height(6000)
thumbnail_slot = rect(w120, h80)


def test_rectangle_is_equal():
    assert rect(w1, h1) == rect(w1, h1)
    assert rect(w1, h1) != rect(w1, h80)
    with pytest.raises(ValueError):
        rect(w1, h1) == "foo"  # noqa


@pytest.mark.parametrize(
    "width, slot, image_format, original_format, expected_filter_str",
    [
        (w53, rect(w120, h80), "jpeg", "jpeg", "width-53"),
        (w53, rect(w120, h80), "jpeg", "avif", "width-53|format-jpeg"),
        (w53, rect(w120, h80), "avif", "jpeg", "width-53|format-avif"),
    ],
)
def test_get_wagtail_filter_str(
    width, slot, image_format: ImageFormat, original_format: ImageFormat, expected_filter_str
):
    rendition_filter = RenditionFilter(width=width, slot=slot, format=image_format)
    assert rendition_filter.get_wagtail_filter_str(original_format) == expected_filter_str


@pytest.mark.parametrize(
    "image, slot, scaled_width",
    [
        (rect(w6000, h4000), rect(w120, h80), 120),  # 3:2 landscape
        (rect(w8000, h4000), rect(w120, h80), 120),  # 2:1 landscape
        (rect(w4000, h6000), rect(w120, h80), 53),  # 2:3 portrait
        (rect(w4000, h4000), rect(w120, h80), 80),  # 1:1 square
    ],
)
def test_calculate_fitting_width(image, slot, scaled_width):
    assert round(calculate_fitting_width(image, slot)) == scaled_width


@pytest.mark.parametrize(
    "image, slot, image_format, max_scale_factor, expected_rendition_filters",
    [
        (rect(w1, h1), rect(w120, h80), "jpeg", 3, []),  # dummy image is too small -> no rendition_filters
        (rect(w1, h1), rect(w120, h80), "avif", 3, []),  # 60 is nearly as big as 53 -> no rendition_filters
        (
            rect(w4000, h6000),
            rect(w120, h80),
            "jpeg",
            3,
            [
                RenditionFilter(width=w53, slot=rect(w120, h80), format="jpeg"),
                RenditionFilter(width=w106, slot=rect(w120, h80), format="jpeg"),
                RenditionFilter(width=w159, slot=rect(w120, h80), format="jpeg"),
            ],
        ),  # 2:3 portrait generates 3 renditions with max_scale_factor=3 because 53*3 < 4000*0.8
    ],
)
def test_get_rendition_filters_for_image_and_slot(
    image, slot, image_format: ImageFormat, max_scale_factor, expected_rendition_filters
):
    rendition_filters = get_rendition_filters_for_image_and_slot(
        image, slot, image_format, max_scale_factor=max_scale_factor
    )
    assert rendition_filters == expected_rendition_filters


@pytest.mark.parametrize(
    "image_name, expected_image_format",
    [
        ("foo.jpg", "jpeg"),
        ("bar.jpeg", "jpeg"),
        ("foo/baz.png", "png"),
        ("an/s/v/g.svg", "svg"),
        ("image.WEBP", "webp"),
        (" image.avif ", "avif"),
    ],
)
def test_get_image_format_by_name(image_name, expected_image_format):
    assert get_image_format_by_name(image_name) == expected_image_format


def test_get_image_format_by_name_not_supported():
    with pytest.raises(ValueError):
        get_image_format_by_name("foo.bmp")


def test_rendition_filters_build_filters():
    # Given a 2:3 portrait image and a thumbnail and a modal image slot
    slots = [
        rect(w120, h80),  # thumbnail
        rect(w1110, h740),  # modal image
    ]
    image = rect(w4000, h6000)  # 2:3 portrait
    # When we get the filters for the image
    filters_dict = RenditionFilters.build_filters(image, slots, image_formats=["avif", "jpeg"])
    filters = RenditionFilters.get_all_filters(filters_dict)
    # Then we get 3 filters for the thumbnail and 3 filters for the modal image
    widths = sorted({f.width for f in filters})
    assert widths == [53, 106, 159, 493, 986, 1479]


def test_rendition_filters_get_filter_by_slot_format_and_fitting_width():
    # Given an empty list of rendition filters
    image_1px = rect(w1, h1)
    [slot] = slots = [Rectangle(w120, h80)]
    rendition_filters = RenditionFilters(
        image=image_1px, original_format="jpeg", slots=slots, image_formats=["avif", "jpeg"]
    )
    # When we try to get a filter by format and width
    with pytest.raises(ValueError):
        # Then we get a ValueError
        rendition_filters.get_filter_by_slot_format_and_fitting_width(slot, "jpeg", w53)

    # Given a list of rendition filters containing a filter for jpeg and width 53
    expected_filter = RenditionFilter(width=w53, slot=thumbnail_slot, format="jpeg")
    rendition_filters.filters[slot]["jpeg"] = [expected_filter]
    # When we get a filter by format and width
    actual_filter = rendition_filters.get_filter_by_slot_format_and_fitting_width(slot, "jpeg", w53)
    # Then we get the expected filter
    assert actual_filter == expected_filter

    # Given there are two filters for jpeg and width 53 and 106
    rendition_filters.filters[slot]["jpeg"] = [expected_filter, expected_filter]
    # When we get a filter by format and width
    with pytest.raises(ValueError):
        # Then we get a ValueError
        rendition_filters.get_filter_by_slot_format_and_fitting_width(slot, "jpeg", w53)



from django.http import HttpRequest

from .models import TemplateBaseDirectory


def site_template_base_dir(request: HttpRequest) -> dict[str, str]:
    """
    Add the name of the template base directory to the context.
    Add the complete base template path to the context for convenience.
    """
    site_template_base_dir_name = TemplateBaseDirectory.for_request(request).name
    return {
        "cast_site_template_base_dir": site_template_base_dir_name,
        "cast_base_template": f"cast/{site_template_base_dir_name}/base.html",
    }



# Generated by Django 4.1.8 on 2023-04-09 06:12

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("cast", "0045_alter_blog_template_base_dir_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="episode",
            name="podcast_audio",
            field=models.ForeignKey(
                blank=True,
                help_text="The audio file for this episode -if this is not set,the episode will not be included in the feed.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="episodes",
                to="cast.audio",
            ),
        ),
    ]



import pytest
from django import forms
from django.http.request import QueryDict
from django.urls import reverse

from cast import appsettings
from cast.models.pages import CustomEpisodeForm, Episode, HomePage, HtmlField, Post
from cast.models.video import Video


class TestVideoModel:
    pytestmark = pytest.mark.django_db

    def test_get_all_video_paths(self, video):
        all_paths = list(video.get_all_paths())
        assert len(all_paths) == 1

    def test_get_all_video_paths_with_poster(self, video_with_poster):
        all_paths = list(video_with_poster.get_all_paths())
        assert len(all_paths) == 2

    def test_get_all_video_paths_without_thumbnail(self, video):
        class Dummy:
            name = "foobar"
            closed = True

            def open(self):
                return None

            def close(self):
                return None

            def seek(self, position):
                return None

            def read(self, num_bytes):
                return b""

            def tell(self):
                return 0

        video.poster = Dummy()
        all_paths = list(video.get_all_paths())
        assert len(all_paths) == 2


class TestGalleryModel:
    @pytest.mark.django_db
    def test_get_image_ids(self, gallery):
        assert len(gallery.image_ids) == gallery.images.count()


class TestFileModel:
    @pytest.mark.django_db
    def test_get_all_file_paths(self, file_instance):
        all_paths = list(file_instance.get_all_paths())
        assert len(all_paths) == 1


class TestBlogModel:
    pytestmark = pytest.mark.django_db

    def test_blog_str(self, blog):
        assert blog.title == str(blog)

    def test_blog_author_null(self, blog):
        blog.author = None
        assert blog.author_name == blog.owner.get_full_name()

    def test_blog_author_not_null(self, blog):
        blog.author = "Foobar"
        assert blog.author_name == blog.author

    def test_paginate_queryset_request_is_none(self, blog):
        context = blog.paginate_queryset({}, blog.get_filterset(QueryDict()).qs, QueryDict())
        assert context["page_obj"].number == 1

    def test_wagtail_api_pages_url(self, blog):
        assert blog.wagtail_api_pages_url == "/cast/api/wagtail/pages/"

    def test_pagination_page_size(self, blog):
        assert blog.pagination_page_size == appsettings.POST_LIST_PAGINATION

    def test_facet_counts_api_url(self, blog):
        assert blog.facet_counts_api_url == reverse("cast:api:facet-counts-detail", kwargs={"pk": blog.pk})

    def test_theme_list_api_url(self, blog):
        assert blog.theme_list_api_url == reverse("cast:api:theme-list")

    def test_theme_update_api_url(self, blog):
        assert blog.theme_update_api_url == reverse("cast:api:theme-update")

    def test_comment_post_url(self, blog):
        assert blog.comment_post_url == reverse("comments-post-comment-ajax")

    def test_has_selectable_themes(self, blog, simple_request):
        assert blog.get_context(simple_request)["has_selectable_themes"]


class TestPostModel:
    pytestmark = pytest.mark.django_db

    def test_post_slug(self, post):
        assert post.get_slug() == "test-entry"

    def test_post_has_audio(self, post):
        assert post.has_audio is False

    def test_episode_has_audio(self, unpublished_episode_without_audio):
        episode = unpublished_episode_without_audio
        assert episode.has_audio is False

    def test_episode_has_audio_true(self, episode, audio):
        episode.podcast_audio = audio
        assert episode.has_audio is True

    def test_post_comments_enabled(self, post, comments_enabled):
        post.comments_enabled = True
        post.blog.comments_enabled = True
        assert post.comments_are_enabled

    def test_post_comments_disabled_settings(self, post, comments_not_enabled):
        post.comments_enabled = True
        post.blog.comments_enabled = True
        assert not post.comments_are_enabled

    def test_post_comments_disabled_blog(self, post, comments_enabled):
        post.comments_enabled = True
        post.blog.comments_enabled = False
        assert not post.comments_are_enabled

    def test_post_comments_disabled_post(self, post, comments_enabled):
        post.comments_enabled = False
        post.blog.comments_enabled = True
        assert not post.comments_are_enabled

    def test_post_media_lookup_value_error(self):
        assert Post().media_lookup == {}

    def test_post_has_audio_value_error(self):
        assert Post().has_audio is False

    def test_media_ids_from_body(self):
        class Block:
            block_type = "image"
            value = None

        class ContentBlock:
            value = (Block(),)

        block = ContentBlock()
        post = Post()
        assert post._media_ids_from_body([block]) == {}

    def test_get_description_escape(self, mocker, simple_request):
        class Rendered:
            rendered_content = "<h1>foo</h1>"

        mocker.patch("cast.models.Post.serve", return_value=Rendered())
        post = Post()
        description = post.get_description(request=simple_request, escape_html=True)
        assert "&lt" in description

    def test_get_description_newlines(self, mocker, simple_request):
        class Rendered:
            rendered_content = "<h1>foo</h1>\n"

        mocker.patch("cast.models.Post.serve", return_value=Rendered())
        post = Post()
        description = post.get_description(request=simple_request, remove_newlines=False)
        assert "\n" in description

    def test_overview_html(self, mocker):
        expected_html = "<h1>foo</h1>"
        mock = mocker.patch("cast.models.Post.get_description", return_value=expected_html)
        html_field = HtmlField(source="*", render_detail=False)
        html_field._context = {"request": "foobar"}
        overview = html_field.to_representation(Post())
        assert overview == expected_html
        assert mock.call_args[1]["render_detail"] is False
        assert mock.call_args[1]["escape_html"] is False
        assert mock.call_args[1]["remove_newlines"] is False

    def test_detail_html(self, mocker):
        expected_html = "<h1>foo</h1><p>bar</p>"
        mock = mocker.patch("cast.models.Post.get_description", return_value=expected_html)
        html_field = HtmlField(source="*", render_detail=True)
        html_field._context = {"request": "foobar"}
        detail = html_field.to_representation(Post())
        assert detail == expected_html
        assert mock.call_args[1]["render_detail"] is True
        assert mock.call_args[1]["escape_html"] is False
        assert mock.call_args[1]["remove_newlines"] is False

    @pytest.mark.parametrize(
        "local_template_name, expected_template",
        [
            (None, "cast/bootstrap4/post.html"),
            ("foobar.html", "cast/bootstrap4/foobar.html"),
        ],
    )
    def test_get_template_for_post(self, local_template_name, expected_template, mocker, simple_request):
        class TemplateBaseDirectory:
            name = "bootstrap4"

        mocker.patch("cast.models.pages.TemplateBaseDirectory.for_request", return_value=TemplateBaseDirectory())
        post = Post()
        post._local_template_name = local_template_name

        assert post.get_template(simple_request) == expected_template

    @pytest.mark.parametrize(
        "is_public, is_removed, contained_in_list",
        [
            (True, False, True),  # public, not removed, in list
            (True, True, False),  # public, removed, not in list
            (False, True, False),  # not public, removed, not in list
            (False, False, False),  # not public, not removed, not in list
        ],
    )
    def test_get_comments(self, is_public, is_removed, contained_in_list, post, comment):
        comment.is_public = is_public
        comment.is_removed = is_removed
        comment.save()
        if contained_in_list:
            [json_comment] = post.comments
            assert json_comment["comment"] == comment.comment
        else:
            assert list(post.comments) == []

    def test_get_comments_is_public_and_is_removed_not_in_fields(self, mocker, post, comment):
        from django_comments import get_model as get_comment_model

        comment_model = get_comment_model()
        exclude = {"is_public", "is_removed"}
        fields_without_excluded = [field for field in comment_model._meta.fields if field.name not in exclude]
        mocker.patch("cast.models.pages.comment_model._meta.fields", fields_without_excluded)
        [json_comment] = post.comments
        assert json_comment["comment"] == comment.comment

    def test_page_type(self):
        post = Post()
        assert post.page_type == "cast.Post"

    def test_podlove_players(self, post_with_audio):
        post = post_with_audio
        [audio] = post.audios.all()
        assert post.podlove_players == [
            (f"#audio_{audio.pk}", audio.podlove_url),
        ]


class TestEpisodeModel:
    pytestmark = pytest.mark.django_db

    def test_get_context_without_absolute_url(self, mocker):
        class Request:
            pass

        mocker.patch("cast.models.Post.get_context", return_value={})
        episode = Episode()
        context = episode.get_context(Request())
        assert "player_url" not in context

    def test_get_enclosure_size_podcast_is_none(self):
        episode = Episode()
        assert episode.get_enclosure_size("mp3") == 0

    @pytest.mark.parametrize(
        "local_template_name, expected_template",
        [
            (None, "cast/bootstrap4/episode.html"),
            ("foobar.html", "cast/bootstrap4/foobar.html"),
        ],
    )
    def test_get_template_for_episode(self, local_template_name, expected_template, mocker, simple_request):
        class TemplateBaseDirectory:
            name = "bootstrap4"

        mocker.patch("cast.models.pages.TemplateBaseDirectory.for_request", return_value=TemplateBaseDirectory())
        episode = Episode()
        episode._local_template_name = local_template_name

        assert episode.get_template(simple_request) == expected_template

    def test_page_type(self):
        episode = Episode()
        assert episode.page_type == "cast.Episode"


@pytest.mark.django_db
def test_custom_episode_form():
    # arrange the custom episode form
    CustomEpisodeForm._meta.model = Episode
    CustomEpisodeForm._meta.fields = ("podcast_audio", "about_to_be_published")
    CustomEpisodeForm.formsets = {}

    # test the form with no data
    form = CustomEpisodeForm()
    assert not form.is_valid()

    # test the form with the "Save draft" button clicked (no audio file is ok)
    form = CustomEpisodeForm({"podcast_audio": None})
    assert form.is_valid()

    # test the form with the "Publish" button clicked (audio file is required)
    form = CustomEpisodeForm({"action-publish": "action-publish", "podcast_audio": None})
    form.fields["podcast_audio"] = forms.IntegerField(required=False)
    assert not form.is_valid()


@pytest.mark.django_db
def test_homepage_serve(episode, mocker):
    mocker.patch("cast.models.pages.Page.serve", return_value="foobar")
    homepage = HomePage()

    # without alias
    assert homepage.serve(None) == "foobar"

    # with alias
    homepage.alias_for_page = episode
    r = homepage.serve(None)
    assert r.status_code == 302


@pytest.mark.django_db
def test_video_create_poster_video_url_without_http(mocker):
    mocker.patch("cast.models.video.tempfile.mkstemp", return_value=(1, "foo"))
    mocker.patch("cast.models.video.Video._get_video_dimensions", return_value=(1, 1))
    check_output = mocker.patch("cast.models.video.check_output", side_effect=ValueError())

    class Original:
        url = "https://example.com/video.mp4"

    video = Video()
    mocker.patch.object(video, "original", Original())
    with pytest.raises(ValueError):
        video._create_poster()
    command = check_output.call_args_list[0][0][0]
    assert "example" in command



import pytest
from django.urls import reverse

from cast.models import Post


class TestPostAdd:
    pytestmark = pytest.mark.django_db

    def test_get_add_form_post_not_authenticated(self, client, blog):
        add_url = reverse("wagtailadmin_pages:add_subpage", args=(blog.id,))
        r = client.get(add_url)

        # redirect to log in
        assert r.status_code == 302
        login_url = reverse("wagtailadmin_login")
        assert login_url in r.url

    def test_get_add_form_post_authenticated(self, client, blog):
        add_url = reverse("wagtailadmin_pages:add_subpage", args=(blog.id,))
        _ = client.login(username=blog.owner.username, password=blog.owner._password)
        r = client.get(add_url, follow=True)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "html" in content
        # make sure we got the wagtail add subpage form and not the login form
        assert '<body id="wagtail" class="  ">' in content

    def test_submit_add_form_post_not_authenticated(self, client, post_data_wagtail, blog):
        add_url = reverse("wagtailadmin_pages:add", args=("cast", "post", blog.id))
        r = client.post(add_url, post_data_wagtail)

        # make sure we are redirected to login page
        assert r.status_code == 302
        login_url = reverse("wagtailadmin_login")
        assert login_url in r.url

    def test_submit_add_form_post_authenticated(self, client, post_data_wagtail, blog):
        _ = client.login(username=blog.owner.username, password=blog.owner._password)
        add_url = reverse("wagtailadmin_pages:add", args=("cast", "post", blog.id))
        r = client.post(add_url, post_data_wagtail)

        # make sure we are redirected to blog index
        assert r.status_code == 302
        assert r.url == reverse("wagtailadmin_explore", args=(blog.id,))

        # make sure there was a post added to the database
        assert Post.objects.get(slug=post_data_wagtail["slug"]).title == post_data_wagtail["title"]

    def test_submit_add_form_post_authenticated_with_image(self, client, post_data_wagtail, blog, image):
        _ = client.login(username=blog.owner.username, password=blog.owner._password)
        add_url = reverse("wagtailadmin_pages:add", args=("cast", "post", blog.id))
        post_data_wagtail["body-0-value-0-type"] = "image"
        post_data_wagtail["body-0-value-0-value"] = image.id

        r = client.post(add_url, post_data_wagtail)

        # make sure we are redirected to blog index
        assert r.status_code == 302
        assert r.url == reverse("wagtailadmin_explore", args=(blog.id,))

        post = Post.objects.get(slug=post_data_wagtail["slug"])

        # make sure there was a post added to the database
        assert post.title == post_data_wagtail["title"]

        # make sure there was an image added
        assert post.images.count() == 1
        assert post.images.first() == image

    def test_submit_add_form_post_authenticated_with_video(self, client, post_data_wagtail, blog, video):
        _ = client.login(username=blog.owner.username, password=blog.owner._password)
        add_url = reverse("wagtailadmin_pages:add", args=("cast", "post", blog.id))
        post_data_wagtail["body-0-value-0-type"] = "video"
        post_data_wagtail["body-0-value-0-value"] = video.id

        r = client.post(add_url, post_data_wagtail)

        # make sure we are redirected to blog index
        assert r.status_code == 302
        assert r.url == reverse("wagtailadmin_explore", args=(blog.id,))

        post = Post.objects.get(slug=post_data_wagtail["slug"])

        # make sure there was a post added to the database
        assert post.title == post_data_wagtail["title"]

        # make sure there was a video added
        assert post.videos.count() == 1
        assert post.videos.first() == video

    def test_submit_add_form_post_authenticated_with_gallery(self, client, post_data_wagtail, blog, gallery):
        _ = client.login(username=blog.owner.username, password=blog.owner._password)
        add_url = reverse("wagtailadmin_pages:add", args=("cast", "post", blog.id))

        post_data_wagtail["body-0-value-0-type"] = "gallery"
        post_data_wagtail["body-0-value-0-value-0-id"] = ""
        post_data_wagtail["body-0-value-0-value-layout"] = "default"
        post_data_wagtail["body-0-value-0-value-gallery-count"] = gallery.images.count()
        post_data_wagtail["body-0-value-0-value-gallery-0-value"] = gallery.images.first().pk
        post_data_wagtail["body-0-value-0-value-gallery-0-deleted"] = ""
        post_data_wagtail["body-0-value-0-value-gallery-0-order"] = "0"

        r = client.post(add_url, post_data_wagtail)

        # make sure we are redirected to blog index
        assert r.status_code == 302
        assert r.url == reverse("wagtailadmin_explore", args=(blog.id,))

        post = Post.objects.get(slug=post_data_wagtail["slug"])

        # make sure there was a post added to the database
        assert post.title == post_data_wagtail["title"]

        # make sure there was a gallery added
        assert post.galleries.count() == 1
        assert post.galleries.first() == gallery
        assert list(post.galleries.first().images.all()) == list(gallery.images.all())

    def test_submit_add_form_post_authenticated_with_code(self, client, post_data_wagtail, blog, video):
        _ = client.login(username=blog.owner.username, password=blog.owner._password)
        add_url = reverse("wagtailadmin_pages:add", args=("cast", "post", blog.id))
        post_data_wagtail["body-0-value-0-type"] = "code"
        post_data_wagtail["body-0-value-0-value-language"] = "python"
        post_data_wagtail["body-0-value-0-value-source"] = 'def hello_world():\n    print("Hello World!")'

        r = client.post(add_url, post_data_wagtail)

        # make sure we are redirected to blog index
        assert r.status_code == 302
        assert r.url == reverse("wagtailadmin_explore", args=(blog.id,))

        post = Post.objects.get(slug=post_data_wagtail["slug"])

        # make sure there was a post added to the database
        assert post.title == post_data_wagtail["title"]

        # make sure there was a code block added
        assert post.body.raw_data[0]["value"][0]["value"]["language"] == "python"
        assert "hello_world" in post.body.raw_data[0]["value"][0]["value"]["source"]
        assert "highlight" in str(post.body)



import logging
import os
import subprocess
import tempfile
from pathlib import Path
from subprocess import check_output
from typing import Optional

from django.contrib.auth import get_user_model
from django.core.files import File as DjangoFile
from django.db import models
from django.utils.translation import gettext_lazy as _
from model_utils.models import TimeStampedModel
from taggit.managers import TaggableManager
from wagtail.models import CollectionMember, PageManager
from wagtail.search import index
from wagtail.search.queryset import SearchableQuerySetMixin

logger = logging.getLogger(__name__)


class VideoQuerySet(SearchableQuerySetMixin, models.QuerySet):
    pass


def get_video_dimensions(lines: list[str]) -> tuple[Optional[int], Optional[int]]:
    """Has its own function to be easier to test."""

    def get_width_height(my_video_type, my_line) -> tuple[int, int]:
        dim_col = my_line.split(", ")[3]
        if my_video_type != "h264":
            dim_col = dim_col.split(" ")[0]
        dim_x, dim_y = tuple(map(int, dim_col.split("x")))
        return dim_x, dim_y

    width, height = None, None
    video_types = ("SAR", "hevc", "h264")
    for line in lines:
        for video_type in video_types:
            if video_type in line:
                width, height = get_width_height(video_type, line)
                break
        else:
            # necessary to break out of nested loop
            continue
        break
    portrait = False
    portrait_triggers = ["rotation of", "DAR 9:16"]
    for line in lines:
        for portrait_trigger in portrait_triggers:
            if portrait_trigger in line:
                portrait = True
    if portrait:
        width, height = height, width
    return width, height


class Video(CollectionMember, index.Indexed, TimeStampedModel):
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE)
    title = models.CharField(default="", max_length=255)
    original = models.FileField(upload_to="cast_videos/")
    poster = models.ImageField(upload_to="cast_videos/poster/", null=True, blank=True)
    poster_seconds = models.FloatField(default=1)

    post_context_key = "video"
    calc_poster = True

    admin_form_fields = ("title", "original", "poster", "tags")
    objects: PageManager["Video"] = VideoQuerySet.as_manager()
    tags = TaggableManager(help_text=None, blank=True, verbose_name=_("tags"))

    search_fields = CollectionMember.search_fields + [
        index.SearchField("title", boost=10),
        index.RelatedFields(
            "tags",
            [
                index.SearchField("name", boost=10),
            ],
        ),
        index.FilterField("user"),
    ]

    @property
    def filename(self) -> str:
        return Path(self.original.name).name

    @property
    def type(self) -> str:
        return "video"

    @staticmethod
    def _get_video_dimensions(video_url: str) -> tuple[Optional[int], Optional[int]]:
        ffprobe_cmd = f'ffprobe -i "{video_url}"'
        result = subprocess.check_output(ffprobe_cmd, shell=True, stderr=subprocess.STDOUT)
        lines = result.decode("utf8").split("\n")
        return get_video_dimensions(lines)

    def _create_poster(self) -> None:
        """Moved into own method to make it mockable in tests."""
        fp, tmp_path = tempfile.mkstemp(prefix="poster_", suffix=".jpg")
        logger.info(f"original url: {self.original.url}")
        video_url = self.original.url
        if not video_url.startswith("http"):
            video_url = self.original.path
        width, height = self._get_video_dimensions(video_url)
        poster_cmd = (
            'ffmpeg -ss {seconds} -i "{video_path}" -vframes 1' " -y -f image2 -s {width}x{height} {poster_path}"
        ).format(
            video_path=video_url,
            seconds=self.poster_seconds,
            poster_path=tmp_path,
            width=width,
            height=height,
        )
        logger.info(poster_cmd)
        check_output(poster_cmd, shell=True)
        name = os.path.basename(tmp_path)
        content = DjangoFile(open(tmp_path, "rb"))
        self.poster.save(name, content, save=False)
        os.unlink(tmp_path)
        logger.info(self.pk)
        logger.info(self.poster)

    def create_poster(self) -> None:
        if self.poster or not self.calc_poster:
            # poster is not null
            logger.info("skip creating poster")
        else:
            try:
                self._create_poster()
            except Exception as e:
                logger.info(e)
                logger.info("create poster failed")

    def get_all_paths(self) -> set[str]:
        paths = set()
        paths.add(self.original.name)
        if self.poster:
            paths.add(self.poster.name)
        return paths

    def get_mime_type(self) -> str:
        ending = self.original.name.split(".")[-1].lower()
        return {
            "mp4": "video/mp4",
            "mov": "video/quicktime",
            "avi": "video/x-msvideo",
        }.get(ending, "video/mp4")

    def save(self, *args, **kwargs) -> Optional["Video"]:
        generate_poster = kwargs.pop("poster", True)
        # need to save original first - django file handling is driving me nuts
        result = super().save(*args, **kwargs)
        if generate_poster:
            logger.info("generate video poster")
            # generate poster thumbnail by default, but make it optional
            # for recalc management command
            self.create_poster()
            # save again after adding poster
            result = super().save(*args, **kwargs)
        return result



#!/usr/bin/env python

import os
import sys


def prepare_notebook_environment():
    from pathlib import Path

    os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"
    os.chdir("notebooks")
    example_project_path = Path(__file__).resolve().parent
    os.environ["PYTHONPATH"] = str(example_project_path)


if __name__ == "__main__":
    # handle DJANGO_SETTINGS_MODULE
    # os.environ.setdefault("DJANGO_SETTINGS_MODULE", "example_site.settings.dev")
    os.environ["DJANGO_SETTINGS_MODULE"] = "example_site.settings.dev"

    # Are we starting a notebook server? -> DJANGO_ALLOW_ASYNC_UNSAFE=true
    # chdir to notebooks and add example project to pythonpath
    if "--notebook" in set(sys.argv):
        prepare_notebook_environment()

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



from .base import *  # noqa


# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "fi6y_c!w#4+16srq_%z+(dj=7d8&5+reik+_171*=e8(0(157x"

# SECURITY WARNING: define the correct hosts in production!
ALLOWED_HOSTS = ["*"]

EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"

INSTALLED_APPS.extend(  # noqa
    [
        "django_extensions",
    ]
)


try:
    from .local import *  # noqa
except ImportError:
    pass



# Generated by Django 3.2.7 on 2021-10-01 09:21

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0020_auto_20210926_1556"),
    ]

    operations = [
        migrations.CreateModel(
            name="SpamFilter",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now, editable=False, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now, editable=False, verbose_name="modified"
                    ),
                ),
                ("name", models.CharField(max_length=128, unique=True)),
                ("model", models.JSONField(default=dict, verbose_name="Spamfilter Model")),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



import logging
from collections import OrderedDict
from typing import Any, cast

from django.contrib.auth.mixins import LoginRequiredMixin
from django.db.models import QuerySet
from django.http import JsonResponse
from django.urls import reverse
from django.views.generic import CreateView
from rest_framework import generics, status
from rest_framework.decorators import api_view
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from wagtail.api.v2.router import WagtailAPIRouter
from wagtail.api.v2.views import PagesAPIViewSet
from wagtail.images.api.v2.views import ImagesAPIViewSet

from ..filters import PostFilterset
from ..forms import SelectThemeForm, VideoForm
from ..models import (
    Audio,
    Blog,
    SpamFilter,
    Video,
    get_template_base_dir,
    get_template_base_dir_choices,
)
from ..views import HtmxHttpRequest
from ..views.theme import set_template_base_dir
from .serializers import (
    AudioPodloveSerializer,
    AudioSerializer,
    FacetCountSerializer,
    SimpleBlogSerializer,
    VideoSerializer,
)
from .viewmixins import AddRequestUserMixin, FileUploadResponseMixin

logger = logging.getLogger(__name__)


@api_view(["GET"])
def api_root(request: Request) -> Response:
    """
    Show API contents.
    If you add any object types, add them here!
    """
    root_api_urls = (
        # ("images", request.build_absolute_uri(reverse("cast:api:image_list"))),
        # ("galleries", request.build_absolute_uri(reverse("cast:api:gallery_list"))),
        ("videos", request.build_absolute_uri(reverse("cast:api:video_list"))),
        ("audios", request.build_absolute_uri(reverse("cast:api:audio_list"))),
        ("comment_training_data", request.build_absolute_uri(reverse("cast:api:comment-training-data"))),
        ("themes", request.build_absolute_uri(reverse("cast:api:theme-list"))),
        ("pages", request.build_absolute_uri(reverse("cast:api:wagtail:pages:listing"))),
        ("images", request.build_absolute_uri(reverse("cast:api:wagtail:images:listing"))),
        ("facet_counts", request.build_absolute_uri(reverse("cast:api:facet-counts-list"))),
    )
    return Response(OrderedDict(root_api_urls))


class VideoCreateView(LoginRequiredMixin, AddRequestUserMixin, FileUploadResponseMixin, CreateView):  # type: ignore
    model = Video
    form_class = VideoForm
    user_field_name = "user"


class StandardResultsSetPagination(PageNumberPagination):
    page_size = 40
    page_size_query_param = "pageSize"
    max_page_size = 10000


class VideoListView(generics.ListCreateAPIView):
    serializer_class = VideoSerializer
    pagination_class = StandardResultsSetPagination
    permission_classes = (IsAuthenticated,)

    def get_queryset(self) -> QuerySet[Video]:
        user = self.request.user
        qs = Video.objects.all().filter(user=user)
        return qs.order_by("-created")


class VideoDetailView(generics.RetrieveDestroyAPIView):
    queryset = Video.objects.all()
    serializer_class = VideoSerializer
    permission_classes = (IsAuthenticated,)


class AudioListView(generics.ListCreateAPIView):
    serializer_class = AudioSerializer
    pagination_class = StandardResultsSetPagination
    permission_classes = (IsAuthenticated,)

    def get_queryset(self) -> QuerySet[Audio]:
        user = self.request.user
        qs = Audio.objects.all().filter(user=user)
        return qs.order_by("-created")


class AudioDetailView(generics.RetrieveDestroyAPIView):
    queryset = Audio.objects.all()
    serializer_class = AudioSerializer
    permission_classes = (IsAuthenticated,)


class AudioPodloveDetailView(generics.RetrieveAPIView):
    queryset = Audio.objects.all()
    serializer_class = AudioPodloveSerializer

    def retrieve(self, request: Request, *args, **kwargs) -> Response:
        instance = self.get_object()
        if (episode_id := request.query_params.get("episode_id")) is not None:
            try:
                episode_id = int(episode_id)
                instance.set_episode_id(episode_id)
            except (ValueError, TypeError):
                pass
        serializer = self.get_serializer(instance)
        return Response(serializer.data)


class FacetCountListView(generics.ListCreateAPIView):
    serializer_class = SimpleBlogSerializer
    queryset = Blog.objects.all().live().order_by("-first_published_at")
    pagination_class = StandardResultsSetPagination


class FacetCountsDetailView(generics.RetrieveAPIView):
    queryset = Blog.objects.all()
    serializer_class = FacetCountSerializer


class CommentTrainingDataView(APIView):
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def get(request, _format: Any = None) -> JsonResponse:
        """
        Return training data for comment classification.
        """
        train = SpamFilter.get_training_data_comments()
        return JsonResponse(train, safe=False)


class ThemeListView(generics.ListAPIView):
    """
    Return a list of available themes. Mark the currently selected theme.
    This is used by the theme switcher for the vue frontend for example.
    """

    def get_queryset(self) -> None:
        return None

    def list(self, request: Request, *args, **kwargs) -> Response:
        choices = get_template_base_dir_choices()
        request = cast(HtmxHttpRequest, request)
        template_base_dir = get_template_base_dir(request, None)
        themes = []
        for slug, name in choices:
            selected = slug == template_base_dir
            themes.append({"slug": slug, "name": name, "selected": selected})
        result = {"items": themes}
        return Response(result)


class UpdateThemeView(APIView):
    """
    Update the selected theme.
    """

    def post(self, request: Request, *args, **kwargs) -> Response:
        new_theme_slug = request.data.get("theme_slug", None)
        form = SelectThemeForm({"template_base_dir": new_theme_slug})
        if not form.is_valid():
            return Response({"error": "Theme slug is invalid"}, status=status.HTTP_400_BAD_REQUEST)

        request = cast(HtmxHttpRequest, request)
        set_template_base_dir(request, form.cleaned_data["template_base_dir"])

        return Response({"message": "Theme updated successfully"}, status=status.HTTP_200_OK)


class FilteredPagesAPIViewSet(PagesAPIViewSet):
    def get_filtered_queryset(self) -> QuerySet:
        # allow additional query parameters from PostFilterset + use_post_filter flag
        additional_query_params = PostFilterset.Meta.fields + ["use_post_filter"] + ["date_before", "date_after"]
        self.known_query_parameters: set = self.known_query_parameters.union(additional_query_params)
        # remove search parameter from query params because it won't work with PagesAPIViewSet
        # in combination with PostFilterset. But doing full text search on PostFilterset will work.
        original_get_params = self.request.GET.copy()
        get_params = self.request.GET.copy()
        if "search" in get_params:
            del get_params["search"]
        self.request.GET = get_params
        queryset = super().get_queryset()
        filterset = PostFilterset(data=original_get_params, queryset=queryset)
        return filterset.qs

    def get_queryset(self):
        if self.request.GET.dict().get("use_post_filter", "false") == "true":
            return self.get_filtered_queryset()
        return super().get_queryset()


# Wagtail API
wagtail_api_router = WagtailAPIRouter("cast:api:wagtail")
wagtail_api_router.register_endpoint("pages", FilteredPagesAPIViewSet)
wagtail_api_router.register_endpoint("images", ImagesAPIViewSet)



# Generated by Django 4.1.5 on 2023-01-28 05:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("cast", "0028_rename_and_drop_itune_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="audio",
            name="data",
            field=models.JSONField(blank=True, default=dict, verbose_name="Metadata"),
        ),
    ]



from django.urls import include, path, re_path
from rest_framework.documentation import include_docs_urls
from wagtail import urls as wagtail_urls
from wagtail.admin import urls as wagtailadmin_urls
from wagtail.documents import urls as wagtaildocs_urls

from cast.views import defaults as default_views_cast

handler404 = default_views_cast.page_not_found
handler500 = default_views_cast.server_error
handler400 = default_views_cast.bad_request
handler403 = default_views_cast.permission_denied


urlpatterns = [
    # rest framework docs/schema urls
    re_path(r"^docs/", include_docs_urls(title="cast API service")),
    path("cast/", include("cast.urls", namespace="cast")),
    # comments
    path("posts/comments/", include("fluent_comments.urls")),
    # wagtail
    path("cms/", include(wagtailadmin_urls)),
    path("documents/", include(wagtaildocs_urls)),
    path("", include(wagtail_urls)),  # default is wagtail
]



"""
Django settings for example_site project.

Generated by 'django-admin startproject' using Django 3.0.3.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.0/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent.parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.0/howto/deployment/checklist/

# GENERAL
# ------------------------------------------------------------------------------
# Local time zone. Choices are
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# though not all of them may be available with every OS.
# In Windows, this must be set to your system time zone.
TIME_ZONE = "Europe/Berlin"
# https://docs.djangoproject.com/en/dev/ref/settings/#language-code
# LANGUAGE_CODE = 'en-us'
LANGUAGE_CODE = "de-DE"
# https://docs.djangoproject.com/en/dev/ref/settings/#site-id
SITE_ID = 1
# https://docs.djangoproject.com/en/dev/ref/settings/#use-i18n
USE_I18N = True
# https://docs.djangoproject.com/en/dev/ref/settings/#use-l10n
# https://docs.djangoproject.com/en/dev/ref/settings/#use-tz
USE_TZ = True

# APPS
# ------------------------------------------------------------------------------
DJANGO_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.admin",
]
THIRD_PARTY_APPS = [
    "crispy_forms",
    "crispy_bootstrap4",
    "django_filters",
    "rest_framework",
    "cast.apps.CastConfig",
    "fluent_comments",
    "threadedcomments",
    "django_comments",
    "wagtail.api.v2",
    "wagtail.contrib.forms",
    "wagtail.contrib.redirects",
    "wagtail.contrib.settings",
    "wagtail.embeds",
    "wagtail.sites",
    "wagtail.users",
    "wagtail.snippets",
    "wagtail.documents",
    "wagtail.images",
    "wagtail.search",
    "wagtail.admin",
    "wagtail",
    "modelcluster",
    "taggit",
]
LOCAL_APPS: list = [
    # Your stuff: custom apps go here
]
# https://docs.djangoproject.com/en/dev/ref/settings/#installed-apps
INSTALLED_APPS = DJANGO_APPS + THIRD_PARTY_APPS + LOCAL_APPS

# MIDDLEWARE
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#middleware
MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "wagtail.contrib.redirects.middleware.RedirectMiddleware",
    "django_htmx.middleware.HtmxMiddleware",
]

ROOT_URLCONF = "example_site.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [BASE_DIR / "example_site" / "templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# See: http://django-crispy-forms.readthedocs.io/en/latest/install.html#template-packs
CRISPY_TEMPLATE_PACK = "bootstrap4"
CRISPY_ALLOWED_TEMPLATE_PACKS = "bootstrap4"

CRISPY_CLASS_CONVERTERS = {
    "textinput": "textinput textInput",
    "fileinput": "fileinput fileUpload",
    "passwordinput": "textinput textInput",
}

WSGI_APPLICATION = "example_site.wsgi.application"


# Database
# https://docs.djangoproject.com/en/3.0/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True


USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/

STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
]

STATICFILES_DIRS = [
    BASE_DIR / "staticfiles",
]

# ManifestStaticFilesStorage is recommended in production, to prevent outdated
# Javascript / CSS assets being served from cache (e.g. after a Wagtail upgrade).
# See https://docs.djangoproject.com/en/3.0/ref/contrib/staticfiles/#manifeststaticfilesstorage
STATICFILES_STORAGE = "django.contrib.staticfiles.storage.ManifestStaticFilesStorage"

STATIC_ROOT = BASE_DIR / "static"
STATIC_URL = "/static/"

MEDIA_ROOT = BASE_DIR / "media"
MEDIA_URL = "/media/"


# Wagtail settings

WAGTAIL_SITE_NAME = "example_site"
WAGTAILIMAGES_MAX_UPLOAD_SIZE = 30 * 1024 * 1024
WAGTAILADMIN_BASE_URL = "cms/"

# Base URL to use when referring to full URLs within the Wagtail admin backend -
# e.g. in notification emails. Don't include '/admin' or a trailing slash
BASE_URL = "http://example.com"

# django admin
ADMIN_URL = "admin/"

# Comments
COMMENTS_APP = "fluent_comments"
FLUENT_COMMENTS_EXCLUDE_FIELDS = ("email", "url", "title")
CAST_COMMENTS_ENABLED = True

# REST
REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.TokenAuthentication",
    ),
    # AttributeError at /api/docs/ -> 'AutoSchema' object has no attribute 'get_link'
    "DEFAULT_SCHEMA_CLASS": "rest_framework.schemas.coreapi.AutoSchema",
}

# wagtail-srcset
# breaks gallery because image renditions are then too small
# SRCSET_DYNAMIC = True

# auto field
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



import shutil
from pathlib import Path

import pytest
from django.template import engines
from django.urls import reverse

from cast.context_processors import site_template_base_dir
from cast.models.theme import (
    get_required_template_names,
    get_template_base_dir_candidates,
    get_template_base_dir_choices,
    get_template_directories,
)


def create_new_theme(name, invalid=False):
    default_engine = list(engines.all())[0]
    first_template_dir = Path(list(list(default_engine.engine.template_loaders)[0].get_dirs())[0])

    new_base_dir = first_template_dir / "cast" / name
    new_base_dir.mkdir(parents=True, exist_ok=True)

    required_names = get_required_template_names()
    if invalid:
        required_names = required_names[:-2]  # remove a required name to make it invalid

    for template_name in required_names:
        path = new_base_dir / template_name
        path.touch()

    return new_base_dir


def test_get_template_base_dir_choices():
    def get_choice_values():
        return {choice[0] for choice in get_template_base_dir_choices()}

    # make sure the theme we create is not already there
    choices = get_choice_values()
    theme_name = "foobar"
    assert theme_name not in choices

    # create an invalid theme
    invalid_name = "invalid"
    invalid_base_dir = create_new_theme(invalid_name, invalid=True)
    choices_with_invalid_theme = get_choice_values()
    assert invalid_name not in choices_with_invalid_theme
    shutil.rmtree(invalid_base_dir)  # cleanup

    # create a valid new theme
    created_base_dir = create_new_theme(theme_name)
    choices_with_new_theme = get_choice_values()
    assert theme_name in choices_with_new_theme

    shutil.rmtree(created_base_dir)  # cleanup
    assert not created_base_dir.exists()


def test_get_template_base_dir_candidates_empty_required_names():
    assert len(get_template_base_dir_candidates([], required_template_names=[])) == 0


def test_get_template_directories_no_compatible_loaders(mocker):
    class FakeLoader:
        pass

    class FakeEngine:
        template_loaders = [FakeLoader()]

    class FakeContainer:
        engine = FakeEngine()

    mocker.patch("cast.models.theme.engines.all", return_value=[FakeContainer()])
    assert get_template_directories() == []


def test_cast_custom_theme_settings_show_up(settings):
    theme_name, theme_display = "my_theme", "My Theme"

    # make sure the custom theme is added to the choices
    settings.CAST_CUSTOM_THEMES = [(theme_name, theme_display)]
    custom_choices = get_template_base_dir_choices()
    assert (theme_name, theme_display) in custom_choices

    # make sure you cannot add predefined themes twice
    settings.CAST_CUSTOM_THEMES.append((theme_name, theme_display))
    assert len(get_template_base_dir_choices()) == len(custom_choices)


@pytest.mark.django_db
def test_context_processors_site_template_base_dir(request_factory):
    request = request_factory.get("/")
    context = site_template_base_dir(request)
    assert context["cast_site_template_base_dir"] == "bootstrap4"
    assert context["cast_base_template"] == "cast/bootstrap4/base.html"


@pytest.mark.django_db
def test_get_select_theme_view(client):
    url = reverse("cast:select-theme")
    response = client.get(url)
    assert response.status_code == 200
    select_template_name = response.templates[0].name
    assert select_template_name == "cast/bootstrap4/select_theme.html"


@pytest.mark.django_db
def test_post_select_theme_view_invalid(client):
    # Given an invalid theme and a next url
    url = reverse("cast:select-theme")
    theme, next_url = "invalid", "/next-url/"
    # When we post to the select theme view
    response = client.post(
        url,
        {
            "template_base_dir": theme,
            "next": next_url,
        },
    )
    # Then we are not redirected to the next url and the theme is not stored in the session
    # and the form was invalid and has an error for the invalid field
    assert response.status_code == 200
    assert "template_base_dir" not in client.session
    assert "template_base_dir" in response.context["theme_form"].errors


@pytest.mark.django_db
def test_post_select_theme_view_happy(client):
    # Given plain as theme and a next url
    url = reverse("cast:select-theme")
    theme, next_url = "plain", "/next-url/"
    # When we post to the select theme view
    response = client.post(
        url,
        {
            "template_base_dir": theme,
            "next": next_url,
        },
    )
    # Then we are redirected to the next url and the theme is stored in the session
    assert response.status_code == 302
    assert next_url == response.url
    assert client.session["template_base_dir"] == "plain"



# Generated by Django 5.0 on 2023-12-23 08:13

import cast.blocks
import wagtail.blocks
import wagtail.embeds.blocks
import wagtail.fields
import wagtail.images.blocks
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cast', '0052_alter_blog_template_base_dir_alter_post_body_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='post',
            name='body',
            field=wagtail.fields.StreamField([('overview', wagtail.blocks.StreamBlock([('heading', wagtail.blocks.CharBlock(form_classname='full title')), ('paragraph', wagtail.blocks.RichTextBlock()), ('code', wagtail.blocks.StructBlock([('language', wagtail.blocks.CharBlock(help_text='The language of the code block')), ('source', wagtail.blocks.TextBlock(help_text='The source code of the block', rows=8))], icon='code')), ('image', cast.blocks.CastImageChooserBlock(template='cast/image/image.html')), ('gallery', wagtail.blocks.StructBlock([('gallery', cast.blocks.GalleryBlock(wagtail.images.blocks.ImageChooserBlock())), ('layout', wagtail.blocks.ChoiceBlock(choices=[('default', 'Web Component with Modal'), ('htmx', 'HTMX based layout')]))])), ('embed', wagtail.embeds.blocks.EmbedBlock()), ('video', cast.blocks.VideoChooserBlock(icon='media', template='cast/video/video.html')), ('audio', cast.blocks.AudioChooserBlock(icon='media', template='cast/audio/audio.html'))])), ('detail', wagtail.blocks.StreamBlock([('heading', wagtail.blocks.CharBlock(form_classname='full title')), ('paragraph', wagtail.blocks.RichTextBlock()), ('code', wagtail.blocks.StructBlock([('language', wagtail.blocks.CharBlock(help_text='The language of the code block')), ('source', wagtail.blocks.TextBlock(help_text='The source code of the block', rows=8))], icon='code')), ('image', cast.blocks.CastImageChooserBlock(template='cast/image/image.html')), ('gallery', wagtail.blocks.StructBlock([('gallery', cast.blocks.GalleryBlock(wagtail.images.blocks.ImageChooserBlock())), ('layout', wagtail.blocks.ChoiceBlock(choices=[('default', 'Web Component with Modal'), ('htmx', 'HTMX based layout')]))])), ('embed', wagtail.embeds.blocks.EmbedBlock()), ('video', cast.blocks.VideoChooserBlock(icon='media', template='cast/video/video.html')), ('audio', cast.blocks.AudioChooserBlock(icon='media', template='cast/audio/audio.html'))]))], use_json_field=True),
        ),
    ]



from django.contrib.auth.models import User
from django.http import HttpRequest

from .htmx_helpers import HtmxHttpRequest


class AuthenticatedHttpRequest(HttpRequest):
    user: User


__all__ = [
    "HtmxHttpRequest",
    "AuthenticatedHttpRequest",
]



import logging
from typing import Literal, Union

from rest_framework import serializers

from ..models import Audio, Blog, Video

logger = logging.getLogger(__name__)


class VideoSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="cast:api:video_detail")
    poster = serializers.ImageField(read_only=True, allow_empty_file=True)
    poster_thumbnail = serializers.ImageField(read_only=True, allow_empty_file=True)

    class Meta:
        model = Video
        fields = ("id", "url", "original", "poster", "poster_thumbnail")


class AudioSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="cast:api:audio_detail")
    podlove = serializers.HyperlinkedIdentityField(view_name="cast:api:audio_podlove_detail")

    class Meta:
        model = Audio
        fields = ("id", "name", "file_formats", "url", "podlove", "mp3")


class AudioPodloveSerializer(serializers.HyperlinkedModelSerializer):
    audio = serializers.ListField()
    chapters = serializers.ListField()
    duration = serializers.CharField(source="duration_str")
    link = serializers.URLField(source="episode_url")

    class Meta:
        model = Audio
        fields = ("title", "subtitle", "audio", "duration", "chapters", "link")


class SimpleBlogSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="cast:api:facet-counts-detail")

    class Meta:
        model = Blog
        fields = ("id", "url")


FacetName = Literal["date_facets", "category_facets", "tag_facets"]
FacetValueName = Literal["slug", "name", "count"]
FacetList = list[dict[FacetValueName, Union[str, int]]]
FacetCounts = dict[FacetName, FacetList]


class FacetCountSerializer(SimpleBlogSerializer):
    facet_counts = serializers.SerializerMethodField()

    class Meta:
        model = Blog
        fields = ("id", "url", "facet_counts")

    def get_facet_counts(self, instance: Blog) -> FacetCounts:
        """
        Facet counts have the following format:
        {
            "facet_name": [
                {
                    "slug": slug,
                    "name": name,
                    "count": count,
                },
            ],
        }
        """
        get_params = self.context["request"].GET.copy()
        filterset = instance.get_filterset(get_params)

        # transform date facets
        date_facets: FacetList = []
        for date, count in filterset.filters["date_facets"].facet_counts.items():
            year_month: str = date.strftime("%Y-%m")
            date_facets.append({"slug": year_month, "name": year_month, "count": count})

        # transform category facets
        category_facets: FacetList = []
        for slug, (name, count) in filterset.filters["category_facets"].facet_counts.items():
            category_facets.append({"slug": slug, "name": name, "count": count})

        # transform tag facets
        tag_facets: FacetList = []
        for slug, (name, count) in filterset.filters["tag_facets"].facet_counts.items():
            tag_facets.append({"slug": slug, "name": name, "count": count})

        return {
            "date_facets": date_facets,
            "category_facets": category_facets,
            "tag_facets": tag_facets,
        }



from unittest.mock import MagicMock

import pytest

from cast.forms import AudioForm, ChapterMarkForm, get_video_form
from cast.models import Audio, ChapterMark


class TestChapterMarkForm:
    pytestmark = pytest.mark.django_db

    def test_add_chapter_mark_form(self, audio):
        row = {"audio": audio.pk, "start": "01:01:01.123", "title": "foo bar baz"}
        form = ChapterMarkForm(row)
        assert form.is_valid()

    def test_add_chapter_mark_form_invalid_url(self, audio):
        row = {
            "audio": audio.pk,
            "start": "01:01:01.123",
            "title": "foo bar baz",
            "link": "foobar",
        }
        form = ChapterMarkForm(row)
        assert not form.is_valid()


class TestAudioForm:
    pytestmark = pytest.mark.django_db

    def test_chaptermarks_not_set(self):
        form = AudioForm({})
        assert form.is_valid()
        assert form.cleaned_data["chaptermarks"] == []

    def test_chaptermarks_empty(self):
        form = AudioForm({"chaptermarks": ""})
        assert form.is_valid()
        assert form.cleaned_data["chaptermarks"] == []

    def test_chaptermarks_invalid_line(self):
        broken_line = "00:12:24.409Dokumentation"
        chaptermarks = "\n".join(
            [
                "00:00:28.433 News aus der Szene",
                "00:02:40.964 Packaging",
                broken_line,
            ]
        )
        form = AudioForm({"chaptermarks": chaptermarks})
        assert form.is_valid() is False
        assert broken_line in form.errors["chaptermarks"][0]

    def test_chaptermarks_invalid_start(self):
        invalid_start = "invalid Dokumentation"
        chaptermarks = "\n".join(
            [
                "00:00:28.433 News aus der Szene",
                "00:02:40.964 Packaging",
                invalid_start,
            ]
        )
        form = AudioForm({"chaptermarks": chaptermarks})
        assert form.is_valid() is False
        assert invalid_start in form.errors["chaptermarks"][0]

    def test_chaptermarks_happy_path(self, user):
        expected_first_title = "News aus der Szene"
        chaptermarks = "\n".join(
            [
                f"00:00:28.433 {expected_first_title}",
                "00:02:40.964 Packaging",
                "00:12:24.409 Dokumentation",
            ]
        )

        # make sure audio form is valid
        audio = Audio(user=user)
        form = AudioForm({"chaptermarks": chaptermarks}, instance=audio)
        assert form.is_valid()
        cleaned_chaptermarks = form.cleaned_data["chaptermarks"]
        assert cleaned_chaptermarks[0].title == expected_first_title

        # make sure chaptermarks are not saved on commit=False
        audio = form.save(commit=False)
        assert ChapterMark.objects.count() == 0

        # make sure chaptermarks are saved in db on form.save()
        audio = form.save(commit=True)
        assert audio.chaptermarks.count() == 3
        assert audio.chaptermarks.order_by("start").first().title == expected_first_title

    def test_old_chaptermarks_are_removed(self, audio):
        ChapterMark.objects.create(audio=audio, start="00:00:28.433", title="old chaptermark")
        assert audio.chaptermarks.count() == 1

        form = AudioForm({}, instance=audio)
        assert form.is_valid()

        # assert old chaptermark was removed
        audio = form.save(commit=True)
        assert audio.chaptermarks.count() == 0

    def test_chaptermarks_from_file(self, audio, m4a_audio):
        expected_title = "News aus der Szene"
        chaptermarks_from_file = [
            {"start": "155.343000", "title": expected_title},
            {"start": "invalid", "title": expected_title},
        ]
        audio.get_chaptermark_data_from_file = MagicMock(return_value=chaptermarks_from_file)
        form = AudioForm({}, {"m4a": m4a_audio}, instance=audio)
        assert form.is_valid()
        assert form.save()

        # make sure chaptermark was saved with correct attributes
        saved_chaptermark = audio.chaptermarks.first()
        assert saved_chaptermark.title == expected_title
        assert str(saved_chaptermark.start) == "00:02:35.343000"  # == "155.343000"


def test_get_video_form_collection_not_added_if_in_admin_form_fields(mocker):
    mocker.patch("cast.forms.Video.admin_form_fields", ("title", "original", "poster", "tags", "collection"))
    video_form = get_video_form()
    assert "collection" in video_form._meta.fields



from django.core.files.storage import InvalidStorageError, Storage

try:
    from django.core.files.storage import storages  # noqa F401

    DJANGO_VERSION_VALID = True
except ImportError:  # pragma: no cover
    DJANGO_VERSION_VALID = False

from django.core.management.base import CommandError

from ...utils import storage_walk_paths


def sync_media_files(source_storage, target_storage):
    for num, path in enumerate(storage_walk_paths(source_storage)):
        if not target_storage.exists(path):
            with source_storage.open(path, "rb") as in_f:
                target_storage.save(path, in_f)
        if num % 100 == 0:  # pragma: no cover
            print(".", end="", flush=True)


def get_production_and_backup_storage() -> tuple[Storage, Storage]:
    if not DJANGO_VERSION_VALID:
        # make sure we run at least Django 4.2
        raise CommandError("Django version >= 4.2 is required")
    else:
        try:
            production_storage, backup_storage = storages["production"], storages["backup"]
            return production_storage, backup_storage
        except InvalidStorageError:
            raise CommandError("production or backup storage not configured")



import pytest

from cast.widgets import AdminAudioChooser, AdminVideoChooser


@pytest.mark.django_db
def test_video_chooser_get_value_data_value_not_none(video):
    avc = AdminVideoChooser()

    # get video data by passing the primary key
    data = avc.get_value_data(video.pk)
    assert data["id"] == video.pk

    # get video data by passing the video object
    data = avc.get_value_data(video)
    assert data["id"] == video.pk


def test_video_chooser_render_js_init():
    avc = AdminVideoChooser()
    js = avc.render_js_init(1, "name", None)
    assert js == "createVideoChooser(1);"


@pytest.mark.django_db
def test_audio_chooser_get_value_data_value_not_none(audio):
    avc = AdminAudioChooser()

    # get audio data by passing the primary key
    data = avc.get_value_data(audio.pk)
    assert data["id"] == audio.pk

    # get audio data by passing the audio object
    data = avc.get_value_data(audio)
    assert data["id"] == audio.pk


def test_audio_chooser_render_js_init():
    avc = AdminAudioChooser()
    js = avc.render_js_init(1, "name", None)
    assert js == "createAudioChooser(1);"



import pytest
from django import http
from django.db import connection
from django.db.utils import DatabaseError

from cast.models import TemplateBaseDirectory
from cast.views import defaults


@pytest.fixture
def my_theme(site):
    name = "my_theme"
    TemplateBaseDirectory.objects.create(name=name, site=site)
    return name


@pytest.mark.django_db
def test_get_template_base_directory_happy(my_theme, simple_request):
    actual_theme = defaults.get_template_base_directory(simple_request)
    assert actual_theme == my_theme


@pytest.mark.django_db
def test_get_template_base_directory_on_database_failure(my_theme, simple_request):
    def blocker(_execute, _sql, _params, _many, _context):
        raise DatabaseError("Simulated broken connection 😱")

    # make sure the default template base directory is used if the database fails
    with connection.execute_wrapper(blocker):
        actual_theme = defaults.get_template_base_directory(simple_request)

    assert actual_theme == "plain"


@pytest.mark.django_db
@pytest.mark.parametrize(
    "view, error_text, expected_response",
    [
        (defaults.page_not_found, "Page not found", http.HttpResponseNotFound),
        (defaults.server_error, "Server Error", http.HttpResponseServerError),
        (defaults.bad_request, "Bad Request", http.HttpResponseBadRequest),
        (defaults.permission_denied, "Permission Denied", http.HttpResponseForbidden),
        (defaults.csrf_failure, "CSRF verification failed.", http.HttpResponseForbidden),
    ],
)
def test_error_handler_views(view, error_text, expected_response, simple_request):
    if error_text == "Server Error":
        response = view(simple_request)
    else:
        response = view(simple_request, Exception())

    # make sure the right response type is returned
    assert isinstance(response, expected_response)

    # make sure the response contains the right error text from the template
    content = response.content.decode("utf-8")
    assert error_text in content



from django.core.management.base import BaseCommand

from .storage_backend import get_production_and_backup_storage, sync_media_files


class Command(BaseCommand):
    help = (
        "backup media files from production to backup storage "
        "(requires Django >= 4.2 and production and backup storage configured)"
    )

    @staticmethod
    def backup_media_files(production_storage, backup_storage):
        sync_media_files(production_storage, backup_storage)

    def handle(self, *args, **options):
        self.backup_media_files(*get_production_and_backup_storage())



from datetime import datetime

import pytest
from django.http import QueryDict
from django.utils.timezone import make_aware

from cast.filters import CountFacetWidget, PostFilterset, SlugChoicesField
from cast.models import Post, PostCategory
from tests.factories import PostFactory


def test_count_facet_widget_render():
    cfw = CountFacetWidget()
    cfw.choices = [("foo", ("bar", "baz"))]
    cfw.data = {}
    html = cfw.render("foo", "bar")  # noqa pycharm warning about template not found
    assert "foo" in html
    assert "bar" in html
    assert "baz" in html


def test_count_facet_widget_if_options(mocker):
    mocker.patch("cast.filters.CountFacetWidget.render_options", return_value=False)
    cfw = CountFacetWidget()
    html = cfw.render("foo", "bar")  # noqa pycharm warning about template not found
    assert "foo" not in html


def test_active_pagination_is_removed_from_count_facet_filter():
    cfw = CountFacetWidget()
    cfw.data = QueryDict("page=3")
    option = cfw.render_option("name", set(), "value", "label")
    assert "page=3" not in option


def test_selected_count_facet_is_in_hidden_input():
    cfw = CountFacetWidget()
    cfw.data = QueryDict("date_facets=2018-12")
    option = cfw.render_option("date_facets", {"2018-12"}, "2018-12", "2018-12 (3)")
    assert '<input type="hidden" name="date_facets" value="2018-12">' in option


@pytest.mark.parametrize(
    "value, is_valid",
    [
        (None, False),
        ("", False),
        ("foo", True),  # happy path
        ("foo bar", False),  # no spaces
    ],
)
def test_validate_category_facet_choice(value, is_valid):
    field = SlugChoicesField()
    assert field.valid_value(value) == is_valid


@pytest.fixture
def post_with_tag(post):
    post.tags.add("tag")
    post.save()
    return post


@pytest.fixture
def post_with_category(post):
    category = PostCategory.objects.create(name="Today I Learned", slug="til")
    post.categories.add(category)
    post.save()  # yes, this is required
    post.category = category
    return post


@pytest.fixture()
def another_post(post, body):
    blog = post.blog
    another_post = PostFactory(owner=blog.owner, parent=blog, title="another post", slug="another-post", body=body)
    another_post.save()
    return another_post


@pytest.mark.django_db
class TestPostFilterset:
    def test_data_is_none(self):
        filterset = PostFilterset(data=None)
        assert filterset.data == QueryDict("")

    def test_queryset_is_none(self):
        filterset = PostFilterset(None, queryset=None)
        assert filterset.qs.count() == 0

    def test_no_posts_no_date_facets(self):
        # given a filterset with no posts
        queryset = Post.objects.none()
        # when the facet counts are fetched
        filterset = PostFilterset(QueryDict(), queryset=queryset)
        # then there are no date facets
        assert filterset.filters["date_facets"].facet_counts == {}

    def test_post_is_counted_in_date_facets(self, post):
        # given a queryset with a post
        queryset = post.blog.unfiltered_published_posts
        # when the facet counts are fetched
        filterset = PostFilterset(QueryDict(), queryset=queryset)
        # then the post is counted in the date facets
        date_facets = filterset.filters["date_facets"].facet_counts
        date_month_post = make_aware(datetime(post.visible_date.year, post.visible_date.month, 1))
        assert date_facets[date_month_post] == 1

    def test_post_is_counted_in_date_facets_when_in_search_result(self, post):
        # given a queryset with a post
        queryset = post.blog.unfiltered_published_posts
        # when the queryset is filtered by the posts title
        querydict = QueryDict(f"search={post.title}")
        filterset = PostFilterset(querydict, queryset=queryset)
        # then the post is in the queryset
        assert post in filterset.qs
        # and the post is counted in the date facets
        date_facets = filterset.filters["date_facets"].facet_counts
        date_month_post = make_aware(datetime(post.visible_date.year, post.visible_date.month, 1))
        assert date_facets[date_month_post] == 1

    def test_post_is_counted_in_date_facets_when_not_in_search_result(self, post):
        # given a queryset with a post
        queryset = post.blog.unfiltered_published_posts
        # when the queryset is filtered by a query that does not match the post
        querydict = QueryDict("search=not_in_title")
        filterset = PostFilterset(querydict, queryset=queryset)
        # then the post is not in the queryset
        assert post not in filterset.qs
        # and the post is not counted in the date facets
        date_facets = filterset.filters["date_facets"].facet_counts
        assert date_facets == {}

    def test_post_is_counted_in_category_facets(self, post_with_category):
        # given a queryset with a post in a category
        post = post_with_category
        queryset = post.blog.unfiltered_published_posts
        # when the facet counts are fetched
        filterset = PostFilterset(QueryDict(), queryset=queryset)
        # then the post is counted in the category facets
        category_facets = filterset.filters["category_facets"].facet_counts
        assert category_facets[post.category.slug] == ("Today I Learned", 1)

    def test_posts_are_filtered_by_category_facet(self, post_with_category, another_post):
        # given a queryset with a post in a category and another post without a category
        post = post_with_category
        blog = post.blog
        # when the posts are filtered by the category
        queryset = blog.unfiltered_published_posts
        querydict = QueryDict("category_facets=til")
        filterset = PostFilterset(querydict, queryset=queryset)
        # then the post without a category is not in the queryset
        assert another_post not in filterset.qs
        assert filterset.qs.count() == 1

    def test_posts_are_filtered_and_wise_by_multiple_categories(self, post, another_post):
        # given a queryset containing two posts being in one category and one of the posts is
        # in an additional category

        # first post
        category = PostCategory.objects.create(name="Today I Learned", slug="til")
        post.categories.add(category)
        another_category = PostCategory.objects.create(name="Additional Category", slug="additional_category")
        post.categories.add(another_category)
        post.save()

        # second post
        blog = post.blog
        another_post.categories.add(category)
        another_post.save()
        queryset = blog.unfiltered_published_posts

        # when the posts are filtered by both categories
        querydict = QueryDict("category_facets=til&category_facets=additional_category")
        filterset = PostFilterset(querydict, queryset=queryset)

        # then the post without the additional category is not in the queryset
        # but the post with both categories is in the queryset
        assert post in filterset.qs
        assert another_post not in filterset.qs
        assert filterset.qs.count() == 1

    def test_posts_are_filtered_by_tag_facet(self, post, another_post):
        # given a queryset with a tagged post and another post without this tag
        post.tags.add("tag")
        post.save()
        # use a tag where name and slug are different to test that the slug is used
        [tag] = post.tags.all()
        tag.slug = "foobar"
        tag.save()
        # when the posts are filtered by the tag
        queryset = post.blog.unfiltered_published_posts
        querydict = QueryDict(f"tag_facets={tag.slug}")
        filterset = PostFilterset(querydict, queryset=queryset)
        # then the untagged post is not in the queryset
        assert another_post not in filterset.qs
        assert filterset.qs.count() == 1

    def test_fields_without_facets_having_some_posts_get_removed(self, post):
        # given a queryset without a post having a category or tag
        queryset = post.blog.unfiltered_published_posts
        # when the facet counts are fetched
        filterset = PostFilterset(QueryDict(), queryset=queryset)
        # then the category and tag facet are not in the filterset form
        assert "category_facets" not in filterset.form.fields
        assert "tag_facets" not in filterset.form.fields

    def test_tag_used_for_two_posts_should_be_counted_twice(self, post, another_post):
        """
        This didn't happen because `PostTag.annotate` created the wrong group
        by clause (post_tag_id instead of tag_id).
        """
        # given two posts with the same tag
        post.tags.add("tag")
        post.save()
        another_post.tags.add("tag")
        another_post.save()
        # when the facet counts are fetched
        queryset = post.blog.unfiltered_published_posts
        filterset = PostFilterset(QueryDict(), queryset=queryset)
        # then the tag is counted twice
        tag_facets = filterset.filters["tag_facets"].facet_counts
        assert tag_facets["tag"] == ("tag", 2)

    def test_bound_field_choices_should_be_set_from_facet_counts(self, post):
        """
        If this test fails, we are in big trouble since there's a lot of magic
        necessary to make the facet counts appear in the choices.
        """
        # given a queryset with a tagged post
        post.tags.add("tag")
        post.save()
        queryset = post.blog.unfiltered_published_posts
        # when we create a filterset
        filterset = PostFilterset(QueryDict(), queryset=queryset)
        # then the tag appears in the choices of the bound field
        slugs = {slug for slug, display in filterset.form["tag_facets"].field.choices}
        assert "tag" in slugs

    def test_remove_filters_not_in_configured_filters(self, mocker):
        # given the configured_filters are an empty set
        mocker.patch("cast.filters.appsettings.CAST_FILTERSET_FACETS", return_value=[])
        # when the filterset is created
        filterset = PostFilterset(QueryDict("foo=bar"))
        # then all filters are removed
        assert filterset.filters == {}



import logging
from typing import TYPE_CHECKING

from django.contrib import admin
from django.contrib.admin import ModelAdmin
from django.db.models import QuerySet

from .models import (
    Audio,
    Blog,
    ChapterMark,
    Episode,
    File,
    Gallery,
    ItunesArtWork,
    Podcast,
    Post,
    SpamFilter,
    Video,
)

if TYPE_CHECKING:
    from django.forms import Form
    from django.http import HttpRequest

logger = logging.getLogger(__name__)


class AdminUserMixin:
    @staticmethod
    def get_changeform_initial_data(request: "HttpRequest") -> dict:
        return {"user": request.user, "author": request.user}


@admin.register(Blog)
class BlogModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("title", "owner")
    fields = ("author", "email", "comments_enabled")


@admin.register(Podcast)
class PodcastModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("title", "owner")
    fields = ("itunes_artwork", "itunes_categories", "explicit", "keywords")


@admin.register(Post)
class PostModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("title", "owner", "blog")
    fields = (
        "visible_date",
        "title",
        "slug",
        "seo_title",
        "search_description",
    )


@admin.register(Episode)
class EpisodeModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("title", "owner", "blog")
    fields = (
        "visible_date",
        "title",
        "slug",
        "seo_title",
        "search_description",
        "podcast_audio",
        "keywords",
        "explicit",
        "block",
    )


@admin.register(ItunesArtWork)
class ItunesArtWorkModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("pk", "original")
    fields = ("original",)


@admin.register(File)
class FileModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("original", "user")
    fields = ("user", "original")


@admin.action(description="Cache file sizes")
def cache_file_sizes(_modeladmin: ModelAdmin, _request: "HttpRequest", queryset: QuerySet[Audio]) -> None:
    for audio in queryset:
        audio.size_to_metadata()
        audio.save()


@admin.register(Audio)
class AudioAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("user", "title", "subtitle", "m4a", "mp3", "oga", "opus")
    fields = ("user", "title", "subtitle", "m4a", "mp3", "oga", "opus", "data")
    actions = [cache_file_sizes]


@admin.register(ChapterMark)
class ChapterMarkModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("start", "title", "link", "image", "audio")


@admin.register(Video)
class VideoModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("pk", "user")

    def save_model(self, request: "HttpRequest", obj: Video, form: "Form", change: bool) -> None:
        logger.info(f"poster: {obj.poster}")
        logger.info(f"form: {form.cleaned_data}")
        print("if in save_model: ", change, not form.cleaned_data["poster"])
        if change and not form.cleaned_data["poster"]:
            logger.info("poster was cleared")
            obj.calc_poster = False
        super().save_model(request, obj, form, change)


@admin.register(Gallery)
class GalleryModelAdmin(AdminUserMixin, ModelAdmin):
    list_display = ("pk",)
    fields = ("layout", "images")


@admin.action(description="Retrain model from scratch using marked comments")
def retrain(_modeladmin: ModelAdmin, _request: "HttpRequest", queryset: QuerySet[SpamFilter]) -> None:
    for spamfilter in queryset:
        train = spamfilter.get_training_data_comments()
        spamfilter.retrain_from_scratch(train)


@admin.register(SpamFilter)
class SpamfilterModelAdmin(ModelAdmin):
    readonly_fields = ["spam", "ham"]
    list_display = tuple(["pk", "name"] + readonly_fields)
    fields = ("name",)
    actions = [retrain]

    @staticmethod
    def spam(obj: SpamFilter) -> dict:
        return obj.performance["spam"]

    @staticmethod
    def ham(obj: SpamFilter) -> dict:
        return obj.performance["ham"]



import pytest
from django.core.files.storage import default_storage

from cast.utils import storage_walk_paths


class TestUtils:
    @pytest.mark.django_db
    def test_walk_fs_paths(self, client, episode):
        audio_path = episode.podcast_audio.m4a.path
        found = False
        for path in storage_walk_paths(default_storage):
            if audio_path.endswith(path):
                found = True
        assert found



import pytest
from django.urls import reverse

from cast.views.gallery import GalleryModalForm, get_prev_next_pk


def test_gallery_modal_form_happy():
    data = {"current_image_pk": 1, "image_pks": "1,2,3", "block_id": "block_id"}
    form = GalleryModalForm(data)
    assert form.is_valid()
    assert form.cleaned_data["image_pks"] == [1, 2, 3]


def test_gallery_modal_form_current_image_not_in_gallery():
    data = {"current_image_pk": 4, "image_pks": "1,2,3", "block_id": "block_id"}
    form = GalleryModalForm(data)
    assert not form.is_valid()
    assert list(form.errors["__all__"])[0] == "current_image_pk 4 is not in image_pks [1, 2, 3]"


def test_gallery_modal_form_current_image_pk_required():
    data = {"image_pks": "1,2,3", "block_id": "block_id"}
    form = GalleryModalForm(data)
    assert not form.is_valid()
    assert form.errors["current_image_pk"][0] == "This field is required."


def test_gallery_modal_form_image_pks_empty():
    data = {"current_image_pk": 4, "image_pks": "", "block_id": "block_id"}
    form = GalleryModalForm(data)
    assert not form.is_valid()
    assert form.errors["image_pks"][0] == "This field is required."


@pytest.mark.parametrize(
    "current_image_pk,image_pks,expected_prev_next",
    [
        (1, [1], (None, None)),
        (1, [1, 2], (None, 2)),
        (1, [1, 2, 3], (None, 2)),
        (2, [1, 2, 3], (1, 3)),
        (3, [1, 2, 3], (2, None)),
    ],
)
def test_get_prev_next_pk(current_image_pk, image_pks, expected_prev_next):
    prev_next = get_prev_next_pk(image_pks, current_image_pk)
    assert prev_next == expected_prev_next


@pytest.mark.django_db
def test_htmx_gallery_modal_happy(client, gallery):
    image_pks = ",".join([str(image.pk) for image in gallery.images.all()])
    current_image_pk = image_pks[0]
    block_id = "block_id"
    base_url = reverse("cast:gallery-modal", kwargs={"template_base_dir": "plain"})
    url = f"{base_url}?current_image_pk={current_image_pk}&image_pks={image_pks}&block_id={block_id}"
    response = client.get(url)
    assert response.status_code == 200


def test_htmx_gallery_modal_without_current_image_pk_invalid(client):
    image_pks = "1,2,3"
    block_id = "block_id"
    base_url = reverse("cast:gallery-modal", kwargs={"template_base_dir": "plain"})
    url = f"{base_url}?&image_pks={image_pks}&block_id={block_id}"
    response = client.get(url)
    assert response.status_code == 400



from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, NewType, cast, get_args

from wagtail.images.models import AbstractImage, AbstractRendition

from . import appsettings

Width = NewType("Width", int)
Height = NewType("Height", int)


@dataclass
class Rectangle:
    """
    Just a simple class to hold width and height. It is hashable to be used as a key
    when getting all renditions for a slot represented by a rectangle.
    """

    width: Width
    height: Height

    def __eq__(self, other):
        if not isinstance(other, Rectangle):
            raise ValueError(f"Can't compare RectDimension with {type(other)}")
        return (self.width, self.height) == (other.width, other.height)

    def __hash__(self):
        return hash((self.width, self.height))


def calculate_fitting_width(image: Rectangle, slot: Rectangle) -> Width:
    """
    Calculate the width of an image that fits into a rectangular slot.

    Returns the width the image needs to be scaled to in order to fit into the rect.
    """

    # Calculate aspect ratios
    image_aspect_ratio = image.width / image.height
    rect_aspect_ratio = slot.width / slot.height

    # Determine if the image needs to be scaled based on width or height
    if image_aspect_ratio > rect_aspect_ratio:
        # Scale based on width
        fitting_width = slot.width
    else:
        # Scale based on height (maintain aspect ratio)
        fitting_width = Width(round(slot.height * image_aspect_ratio))

    return fitting_width


ImageType = Literal["regular", "gallery"]
ImageFormat = Literal["jpeg", "avif", "webp", "png", "svg"]
SUPPORTED_IMAGE_FORMATS = set(get_args(ImageFormat))
ImageFormats = Iterable[ImageFormat]
IMAGE_TYPE_TO_SLOTS: dict[ImageType, list[Rectangle]] = {
    "regular": [Rectangle(Width(w), Height(h)) for w, h in appsettings.CAST_REGULAR_IMAGE_SLOT_DIMENSIONS],
    "gallery": [Rectangle(Width(w), Height(h)) for w, h in appsettings.CAST_GALLERY_IMAGE_SLOT_DIMENSIONS],
}
DEFAULT_IMAGE_FORMATS = cast(ImageFormats, appsettings.CAST_IMAGE_FORMATS)


@dataclass
class RenditionFilter:
    """
    A rendition filter for an image that fits into a slot.
    """

    width: Width  # width of the rendition
    slot: Rectangle  # slot the image needs to fit into
    format: ImageFormat  # desired image format

    def get_wagtail_filter_str(self, original_format: ImageFormat) -> str:
        """Return the filter string in wagtail format."""
        filter_parts = [f"width-{self.width}"]
        if self.format != original_format:
            filter_parts.append(f"format-{self.format}")
        return "|".join(filter_parts)


def get_rendition_filters_for_image_and_slot(
    image: Rectangle,  # dimensions of the original image
    slot: Rectangle,  # slot the image needs to fit into
    image_format: ImageFormat,  # desired image format
    max_scale_factor: int = 3,  # don't scale up renditions more than this
) -> list[RenditionFilter]:
    """
    Get a list of rendition filters for an image that has to fit into a slot.
    Don't scale up renditions more than max_scale_factor. If the rendition_width
    is nearly as big as the image_width, don't create an additional rendition filter.
    """
    filters = []
    fitting_width = calculate_fitting_width(image, slot)
    for pixel_density in range(1, max_scale_factor + 1):
        rendition_width = Width(fitting_width * pixel_density)
        if rendition_width > image.width * 0.8:
            # already big enough
            continue
        filters.append(RenditionFilter(width=rendition_width, slot=slot, format=image_format))
    return filters


def get_image_format_by_name(file_name: str) -> ImageFormat:
    """Guess the image format from the file name."""
    suffix_to_format: dict[str, ImageFormat] = {
        "jpg": "jpeg",
    }
    suffix = Path(file_name).suffix.lower().strip().lstrip(".")
    suffix = suffix_to_format.get(suffix, suffix)
    if suffix not in SUPPORTED_IMAGE_FORMATS:
        raise ValueError(f"Image format {suffix} not supported.")
    else:
        return cast(ImageFormat, suffix)


class ImageForSlot:
    """
    Image fitting into a slot. It has a src and a srcset for all image formats.
    Its purpose is to be used in a <picture> or <img> tag in a django template.
    It is potentially a lot smaller than the slot.

    For example if there's a slot of 120x80 and an image of 4000x6000, the image
    for the slot is 53x80. The srcset contains renditions of 53, 106 and 159 pixels.
    """

    def __init__(
        self,
        image: Rectangle,
        src: dict[ImageFormat, str],
        srcset: dict[ImageFormat, str],
    ) -> None:
        self.width = image.width
        self.height = image.height
        self.sizes = f"{self.width}px"
        self.src = src
        self.srcset = srcset


Filters = dict[Rectangle, dict[ImageFormat, list[RenditionFilter]]]


class RenditionFilters:
    def __init__(
        self,
        *,
        image: Rectangle,
        original_format: ImageFormat,
        slots: list[Rectangle],
        image_formats: ImageFormats,
    ) -> None:
        super().__init__()
        self.image = image
        self.original_format = original_format
        self.image_formats = image_formats
        self.slots = slots
        self.slot_to_fitting_width: dict[Rectangle, Width] = {}
        for slot in slots:
            self.slot_to_fitting_width[slot] = Width(calculate_fitting_width(image, slot))
        self.filters = self.build_filters(self.image, self.slots, self.image_formats)
        self.filter_to_url: dict[str, str] = {}

    @classmethod
    def from_wagtail_image(cls, image: AbstractImage, slots: list[Rectangle], image_formats: ImageFormats):
        original_format = get_image_format_by_name(image.file.name)
        image = Rectangle(Width(image.width), Height(image.height))
        return cls(image=image, original_format=original_format, slots=slots, image_formats=image_formats)

    @classmethod
    def from_wagtail_image_with_type(cls, image: AbstractImage, image_type: ImageType):
        return cls.from_wagtail_image(
            image, slots=IMAGE_TYPE_TO_SLOTS[image_type], image_formats=DEFAULT_IMAGE_FORMATS
        )

    def set_filter_to_url_via_wagtail_renditions(self, renditions: dict[str, AbstractRendition]) -> None:
        self.filter_to_url = {fs: renditions[fs].url for fs in self.filter_strings}

    @staticmethod
    def build_filters(image: Rectangle, slots: list[Rectangle], image_formats: ImageFormats) -> Filters:
        """
        Build all filters for all slots and image formats.
        """
        filters: Filters = {slot: {} for slot in slots}
        for slot in slots:
            for image_format in image_formats:
                filters[slot][image_format] = get_rendition_filters_for_image_and_slot(image, slot, image_format)
        return filters

    def get_filter_by_slot_format_and_fitting_width(
        self, slot: Rectangle, image_format: ImageFormat, fitting_width: Width
    ) -> RenditionFilter:
        """
        Get a rendition filter by format and width. Raise ValueError if no filter found
        or more than one filter found.
        """
        filters = [f for f in self.filters[slot][image_format] if f.width == fitting_width]
        if len(filters) == 0:
            raise ValueError(f"No filter found for format {image_format} and width {fitting_width}")
        if len(filters) > 1:
            raise ValueError(f"More than one filter found for format {image_format} and width {fitting_width}")
        return filters[0]

    @staticmethod
    def get_all_filters(filters: dict[Rectangle, dict[ImageFormat, list[RenditionFilter]]]) -> list[RenditionFilter]:
        """
        Return a flat list of all filters.
        """
        all_filters = []
        for slot_filters in filters.values():
            for format_filters in slot_filters.values():
                all_filters.extend(format_filters)
        return all_filters

    @property
    def all_filters(self) -> list[RenditionFilter]:
        return self.get_all_filters(self.filters)

    def get_filter_strings(self, original_format: ImageFormat) -> list[str]:
        """
        Return a list of filter strings in wagtail format.
        """
        return [f.get_wagtail_filter_str(original_format) for f in self.all_filters]

    @property
    def filter_strings(self) -> list[str]:
        """
        Return a list of filter strings in wagtail format.
        """
        return self.get_filter_strings(self.original_format)

    def get_src_for_slot(self, slot: Rectangle) -> dict[ImageFormat, str]:
        src = {}
        fitting_width = self.slot_to_fitting_width[slot]
        for image_format in self.image_formats:
            format_filter = self.get_filter_by_slot_format_and_fitting_width(slot, image_format, fitting_width)
            format_filter_string = format_filter.get_wagtail_filter_str(self.original_format)
            src[image_format] = self.filter_to_url[format_filter_string]
        return src

    def get_srcset_for_slot(self, slot: Rectangle) -> dict[ImageFormat, str]:
        srcset = {}
        filters_for_slot = self.filters[slot]
        for image_format in self.image_formats:
            filters_for_format = filters_for_slot[image_format]
            filter_strings_for_format = [f.get_wagtail_filter_str(self.original_format) for f in filters_for_format]
            urls_for_filters = [self.filter_to_url[fs] for fs in filter_strings_for_format]
            srcset_parts = []
            for url, filter_string in zip(urls_for_filters, filters_for_format):
                srcset_parts.append(f"{url} {filter_string.width}w")
            srcset[image_format] = ", ".join(srcset_parts)
        return srcset

    def get_image_for_slot(self, slot: Rectangle) -> ImageForSlot:
        src = self.get_src_for_slot(slot)
        srcset = self.get_srcset_for_slot(slot)
        fitting_image = Rectangle(width=self.slot_to_fitting_width[slot], height=slot.height)
        return ImageForSlot(image=fitting_image, src=src, srcset=srcset)



from typing import cast

import pytest
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import AbstractImage, AbstractRendition

from cast.blocks import (
    CastImageChooserBlock,
    CodeBlock,
    GalleryBlock,
    GalleryBlockWithLayout,
    get_srcset_images_for_slots,
)
from cast.models import Gallery
from cast.renditions import IMAGE_TYPE_TO_SLOTS, Height, Rectangle, Width


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, ""),  # make sure None is rendered as empty string
        (  # make sure source is rendered even if language is not found
            {"language": "nonexistent", "source": "blub"},
            '<div class="highlight"><pre><span></span>blub\n' "</pre></div>\n",
        ),
        (  # happy path
            {
                "language": "python",
                "source": "print('hello world!')",
            },
            (
                '<div class="highlight"><pre><span></span><span class="nb">print'
                '</span><span class="p">(</span><span class="s1">&#39;hello world!&#39;'
                '</span><span class="p">)</span>\n'
                "</pre></div>\n"
            ),
        ),
    ],
)
def test_code_block_value(value, expected):
    block = CodeBlock()
    rendered = block.render_basic(value)
    assert rendered == expected


@pytest.mark.parametrize(
    "context",
    [
        None,  # context is None
        {},  # template_base_dir not in context
        {"template_base_dir": "nonexistent"},  # template_base_dir does not exist
    ],
)
def test_gallery_block_template_default(context):
    block = GalleryBlock(ImageChooserBlock())
    assert block.get_template(images=None, context=context) == block.meta.template


def test_gallery_block_template_from_theme(mocker):
    mocker.patch("cast.blocks.get_template")
    block = GalleryBlock(ImageChooserBlock())
    template_name = block.get_template(context={"template_base_dir": "vue"})
    assert template_name == "cast/vue/gallery.html"


class StubWagtailImage:
    class File:
        name = "test.jpg"
        url = "https://example.com/test.jpg"

    width = Width(6000)
    height = Height(4000)
    file = File()

    @staticmethod
    def get_renditions(*filter_strings):
        class StubImage:
            url = "https://example.com/test.jpg"
            width = Width(120)

        return {fs: StubImage() for fs in filter_strings}


def test_get_srcset_images_for_slots_get_renditions_is_called_when_filters_not_empty():
    # Given an image that should generate multiple renditions for a slot
    slot = Rectangle(Width(120), Height(80))
    images_for_slots = get_srcset_images_for_slots(cast(AbstractImage, StubWagtailImage()), "gallery")
    # When we get the srcset images for the slot
    image_for_slot = images_for_slots[slot]
    split_srcset = image_for_slot.srcset["jpeg"].replace(",", "").split(" ")  # type: ignore
    srcset_widths = sorted([int(t.rstrip("w")) for t in split_srcset if t.endswith("w")])
    # Then the srcset widths should be 120 * 1, 120 * 2, 120 * 3
    assert srcset_widths == [120, 240, 360]


class Stub1PxImage:
    class File:
        name = "test.jpg"
        url = "https://example.com/test.jpg"

    width = Width(1)
    height = Height(1)
    file = File()

    @staticmethod
    def get_rendition(_filter_string):
        class Avif:
            url = "https://example.com/test.avif"
            width = Width(1)

        return Avif()


def test_get_srcset_images_for_slots_use_original_if_image_too_small():
    # Given an image that is too small for the slot
    slot = IMAGE_TYPE_TO_SLOTS["gallery"][0]
    images_for_slots = get_srcset_images_for_slots(cast(AbstractImage, Stub1PxImage()), "gallery")
    # When we get the srcset images for the slot
    image_for_slot = images_for_slots[slot]
    # Then the image for the slot should be the original image
    assert image_for_slot.src["jpeg"] == "https://example.com/test.jpg"
    assert image_for_slot.srcset["jpeg"] == "https://example.com/test.jpg 1w"
    # And it should have been converted to avif
    assert image_for_slot.src["avif"] == "https://example.com/test.avif"
    assert image_for_slot.srcset["avif"] == "https://example.com/test.avif 1w"


class StubBigImage:
    class File:
        name = "test.jpg"
        url = "https://example.com/test.jpg"

    url = "https://example.com/test.jpg"
    width = Width(6000)
    height = Height(4000)
    file = File()

    @staticmethod
    def get_renditions(*filter_strings):
        class StubImage:
            url = "https://example.com/test.jpg"
            width = Width(120)

        return {fs: StubImage() for fs in filter_strings}


def test_get_srcset_images_for_slots_fetched_renditions_not_none():
    [slot] = IMAGE_TYPE_TO_SLOTS["regular"]
    rendition = cast(AbstractRendition, StubBigImage())
    image = cast(AbstractImage, StubBigImage())
    images_for_slot = get_srcset_images_for_slots(image, "regular", fetched_renditions={"width-1110": rendition})
    assert images_for_slot[slot].src["jpeg"] == "https://example.com/test.jpg"


def test_get_srcset_images_for_slots_fetched_renditions_contain_all_filter_strings():
    [slot] = IMAGE_TYPE_TO_SLOTS["regular"]
    rendition = cast(AbstractRendition, StubBigImage())
    image = cast(AbstractImage, StubBigImage())
    all_filter_strings = [
        "width-1110",
        "width-2220",
        "width-3330",
        "width-1110|format-avif",
        "width-2220|format-avif",
        "width-3330|format-avif",
    ]
    fetched_renditions = {fs: rendition for fs in all_filter_strings}
    images_for_slot = get_srcset_images_for_slots(image, "regular", fetched_renditions=fetched_renditions)
    assert images_for_slot[slot].src["jpeg"] == "https://example.com/test.jpg"


def test_image_chooser_block_get_context_parent_context_none():
    """Just make sure parent context is set to {} if it is None."""
    cicb = CastImageChooserBlock()
    context = cicb.get_context(StubWagtailImage(), parent_context=None)
    assert "value" in context


def test_gallery_block_get_context_parent_context_none():
    """Just make sure parent context is set to {} if it is None."""
    cb = GalleryBlock(ImageChooserBlock())
    context = cb.get_context(Gallery.objects.none(), parent_context=None)
    assert "value" in context


def test_gallery_block_with_layout_get_template_htmx():
    block = GalleryBlockWithLayout()
    template = block.get_template({"layout": "htmx"}, context={"template_base_dir": "bootstrap4"})
    assert template == "cast/bootstrap4/gallery_htmx.html"


def test_gallery_block_with_layout_get_template_value_is_none():
    block = GalleryBlockWithLayout()
    template = block.get_template(None, context={"template_base_dir": "bootstrap4"})
    assert template == "cast/bootstrap4/gallery.html"

    template = block.get_template({}, context={"template_base_dir": "bootstrap4"})
    assert template == "cast/bootstrap4/gallery.html"



from typing import Any

from django.http import HttpRequest
from fluent_comments.models import FluentComment
from fluent_comments.moderation import FluentCommentsModerator

from .models import SpamFilter


class Moderator(FluentCommentsModerator):
    def __init__(self, model, spamfilter=None):
        super().__init__(model)
        # Allow spamfilter to be set for tests
        if spamfilter is not None:
            self.spamfilter = spamfilter
        else:
            self.spamfilter = SpamFilter.get_default()

    def allow(self, comment: FluentComment, content_object: Any, request: HttpRequest) -> bool:
        """
        Allow all. Just mark moderated comments as 'is_removed' but
        keep them in the database. Even awful comments are useful as
        a bad training example :).
        """
        return True

    def moderate(self, comment: FluentComment, content_object: Any, request: HttpRequest) -> bool:
        message = SpamFilter.comment_to_message(comment)
        if self.spamfilter is not None:
            predicted_label = self.spamfilter.model.predict_label(message)
        else:
            predicted_label = "unknown"
        if predicted_label == "spam":
            comment.is_removed, comment.is_public = True, False
            return True
        else:
            comment.is_removed, comment.is_public = False, True
            return False
